/*
**
** 3D.C        Does a 3D plot of the CCD data
**
** Will Menninger, 10/89, programmed in Turbo C 2.0
**
*/

#include "contour.h"

#define TM      .1
#define TM2D    .1
#define BM      .05
#define BM2D    .1
#define RM      .05
#define RM2D    .05
#define LM      .15
#define LM2D    .3
#define LEGMAR  .07
#define MAINX   .6
#define MAINY   .6
#define VFRACB  .65

/* colors */
#define BACKGRND        0
#define AXES            1
#define AXESBRIGHT      15
#define NLAYERCOLORS    4
#define MAXLAYERS       80
static int palette3Dc[16][3]=
        {
            {0x40,0x40,0x40}, /* Background */
            {0x90,0x90,0x90}, /* Axes */
            {0x00,0xff,0x00}, /* Green top fill   */
            {0x00,0x80,0x00}, /* Green bot fill   */
            {0x00,0x40,0x00}, /* Green wire frame */
            {0x80,0x00,0xff}, /* Blue-Purple top fill */
            {0x40,0x00,0x80}, /* Blue-Purple bot fill */
            {0x20,0x00,0x40}, /* Blue-Purple wire frame */
            {0xff,0x00,0x00}, /* Red top fill */
            {0x80,0x00,0x00}, /* Red bot fill */
            {0x40,0x00,0x00}, /* Red wire frame */
            {0xff,0xc0,0x00}, /* Orange-Yellow top fill */
            {0x80,0x60,0x00}, /* Orange-Yellow bot fill */
            {0x40,0x30,0x00}, /* Orange-Yellow wire frame */
            {0xff,0xff,0x00}, /* Unused */
            {0xff,0xff,0xff}  /* Bright White */
        };
static int palette3Dg[16][3]=
        {
            {0xff,0xff,0xff}, /* Background */
            {0x00,0x00,0x00}, /* Axes */
            {0xd0,0xd0,0xd0}, /* Green top fill   */
            {0xd0,0xd0,0xd0}, /* Green bot fill   */
            {0x00,0x00,0x00}, /* Green wire frame */
            {0xa0,0xb0,0xa0}, /* Blue-Purple top fill */
            {0xa0,0xb0,0xa0}, /* Blue-Purple bot fill */
            {0x00,0x00,0x00}, /* Blue-Purple wire frame */
            {0x70,0x70,0x70}, /* Red top fill */
            {0x70,0x70,0x70}, /* Red bot fill */
            {0x00,0x00,0x00}, /* Red wire frame */
            {0x50,0x50,0x50}, /* Orange-Yellow top fill */
            {0x50,0x50,0x50}, /* Orange-Yellow bot fill */
            {0x00,0x00,0x00}, /* Orange-Yellow wire frame */
            {0x00,0x00,0x00}, /* Unused */
            {0x00,0x00,0x00}  /* Bright White */
        };
static int layerpen[4][3] = {{2,3,4},{5,6,7},{8,9,10},{11,12,13}};
static int nlayers;
static int startlayer;
float   zboundary[MAXLAYERS+1];

static int  xoff;
static int  yoff;
static int  xscale;
static int  yscale;
static float    zscale;
static float    zmin;
static float    gyx,gyy,gyz,gy0;
static float    gxx,gxy,gx0;
static int  diagonal;
static int viewtype;

static void setup3d    (int viewtype,DATAPTR dp);
static int  x3d        (float x,float y);
static int  y3d        (float x,float y,float z);
static void line3d     (float x1,float y1,float z1,float x2,float y2,float z2);
static void setupaxes  (DATAPTR d,int grids);
static void triangle3d (float x1,float y1,float z1,float x2,float y2,
                        float z2,float x3,float y3,float z3);
static void tile3d     (float x1,float y1,float z1,float x2,float y2,float z2,
                        float x3,float y3,float z3,float x4,float y4,float z4);
static void plot3d     (float far *far*sd,float far *far*in,
                        DATAPTR dp,int grids);
static void colors3d   (int grey_scale);
static void top_pen    (int layer);
static void bot_pen    (int layer);
static void wire_pen   (int layer);
static void three_verts(int grids,float z0,float z1);
static void fix_boundaries(void);
static void heapsort(void);
static void do_levels(char *buf,DATAPTR d);
static void add_boundaries(float start,float end,float step);
static void z_legend(DATAPTR d);


void do_3d(DATAPTR d)

    {
    char    buf[100];
    char    bvals[100];
    float   xmin,xmax,ymin,ymax,zmax;
    int     key;
    int     grids;

    if (!loaded(TRUE))
        return;
    zscale=d->max;
    while (1)
        {
        if (!get_range('X',d,&xmin,&xmax))
            return;
        xoff=xindex(d,xmin)+.0001;
        xscale=xindex(d,xmax)+.99-xoff;
        if (xoff<0)
            xoff=0;
        if (xoff>d->xsize-2)
            xoff=d->xsize-2;
        if (xoff+xscale>d->xsize-1)
            xscale=d->xsize-1-xoff;
        if (xscale>0)
            break;
        error("Specified range is too small.");
        normal("\r\n\r\n");
        }
    while (1)
        {
        if (!get_range('Y',d,&ymin,&ymax))
            return;
        yoff=yindex(d,ymin)+.0001;
        yscale=yindex(d,ymax)+.99-yoff;
        if (yoff<0)
            yoff=0;
        if (yoff>d->ysize-2)
            yoff=d->ysize-2;
        if (yoff+yscale>d->ysize-1)
            yscale=d->ysize-1-yoff;
        if (yscale>0)
            break;
        error("Specified range is too small.");
        normal("\r\n\r\n");
        }
    /*
        if (bgetinput("Enter CX,CY,XRADIUS,YRADIUS:  ",buf,40)==ESCAPE)
            return;
        normal("\r\n\r\n");
        if (buf[0]=='\0')
            i=0;
        else
            i=sscanf(buf,"%f,%f,%f,%f",&cx,&cy,&xr,&yr);
        if (i==3)
            yr=xr;
        if (i<3)
            {
            xr=(d->xmax-d->xmin)/2.;
            yr=(d->ymax-d->ymin)/2.;
            }
        if (i<2)
            cy=(d->ymax+d->ymin)/2.;
        if (i<1)
            cx=(d->xmax+d->xmin)/2.;
        xoff=xindex(d,xmin);
        yoff=yindex(d,ymin);
        xscale=xindex(d,xmax)+.99-xoff;
        yscale=yindex(d,ymax)+.99-yoff;
        if (xoff<0)
            xoff=0;
        if (yoff<0)
            yoff=0;
        if (xoff>d->xsize-2)
            xoff=d->xsize-2;
        if (yoff>d->ysize-2)
            yoff=d->ysize-2;
        if (xoff+xscale>d->xsize-1)
            xscale=d->xsize-1-xoff;
        if (yoff+yscale>d->ysize-1)
            yscale=d->ysize-1-yoff;
        if (xscale<=0 || yscale<=0)
            {
            error("Radii must be larger.");
            normal("\r\n\r\n");
            continue;
            }
        break;
        }
    */
    if (!get_range('Z',d,&zmin,&zmax))
        return;
    /*
    if (bgetinput("Enter Zmin,Zmax:  ",buf,40)==ESCAPE)
        return;
    normal("\r\n\r\n");
    if (buf[0]=='\0')
        i=0;
    else
        i=sscanf(buf,"%f,%f",&zmin,&zmax);
    if (i<2)
        zmax=d->max;
    if (i<1)
        zmin=d->min>=0. ? 0. : d->min;
    */
    zscale=zmax-zmin;
    normal("Enter layer boundary values.  Prefix a \"/\" for step values.\r\n");
    normal("E.g. \"1.5 3 6 /2 12\" will do values 1.5, 3, 6, 8, 10, and 12.\r\n");
    sprintf(buf,"Values (%d max):  ",MAXLAYERS);
    if (bgetinput(buf,bvals,55)==ESCAPE)
        return;
    normal("\r\n\r\n");
    normal("Select View:  <A> zero degree   <B> 45 degree   <C> overhead\r\n");
    while (1)
        {
        key=bbioskey(0);
        key=tolower(key&0xff);
        if (key=='a' || key==13 || key==27 || key=='b' || key=='c')
            break;
        }
    if (key==27)
        return;
    if (key==13)
        key='a';
    viewtype=key;
    if (viewtype!='c')
        {
        normal("\r\nGrids?  (Y/N)\r\n");
        while (1)
            {
            key=bbioskey(0);
            key=tolower(key&0xff);
            if (key=='y' || key==13 || key==27 || key=='n')
                break;
            }
        if (key==27)
            return;
        if (key==13)
            key='y';
        grids=(key=='y');
        }
    else
        grids=0;
    do_levels(bvals,d);
    fix_boundaries();
    plot3d(d->shotdata,d->interp,d,grids);
    }


static void fix_boundaries(void)

    {
    int     i,j;

    /* Remove boundaries outside (zmin,zmax) */
    startlayer=0;
    for (i=j=0;i<=nlayers;i++)
        {
        if (zboundary[i]<zmin)
            {
            startlayer++;
            continue;
            }
        if (zboundary[i]<zmin || zboundary[i]>zmin+zscale)
            continue;
        zboundary[j++]=zboundary[i];
        }
    nlayers=j-1;

    /* Sort */
    heapsort();

    /* Remove double points */
    for (i=1,j=0;i<=nlayers;i++)
        {
        if ((zboundary[i]-zboundary[j])/zscale < 0.001)
            continue;
        zboundary[++j]=zboundary[i];
        }
    nlayers=j;
    }


/*
** Heapsort algorithm, from Numerical Recipes in C, p. 247
** Used to sort the lookup table in the index array based on
** the variable names in the list array.
** Converted to use indices 0 to n-1.
*/

static void heapsort(void)

    {
    int     l,j,ir,i,n;
    float   x0;

    n=nlayers+1;
    l=n>>1;
    ir=n-1;
    while (1)
        {
        if (l>0)
            {
            l--;
            x0=zboundary[l];
            }
        else
            {
            x0=zboundary[ir];
            zboundary[ir]=zboundary[0];
            ir--;
            if (!ir)
                {
                zboundary[0]=x0;
                return;
                }
            }
        i=l;
        j=((l+1)<<1)-1;
        while (j<=ir)
            {
            if (j<ir && zboundary[j]<zboundary[j+1])
                j++;
            if (x0<zboundary[j])
                {
                zboundary[i]=zboundary[j];
                i=j;
                j+=(i+1);
                }
            else
                j=ir+1;
            }
        zboundary[i]=x0;
        }
    }


static void do_levels(char *buf,DATAPTR d)

    {
    int     i,j,i1,i2,modval,ls,notstarted,stepval;
    float   start,z,zmax,val,fstep,offset,scale;
    char    buf2[100];

    zmax=zmin+zscale;
    zboundary[0]=zmin;
    zboundary[1]=zmax;
    nlayers=1;
    if (buf[0]==EOS)
        {
        getticks(zmin,zmin+zscale,&offset,&scale,&i1,&i2,&modval,
                 viewtype=='c'?30:20);
        for (i=i1+1;i<=i2;i++)
            {
            if (i%modval)
                continue;
            z=offset+i*scale;
            if (fabs(z/scale)<1e-5)
                z=0;
            nlayers++;
            zboundary[nlayers]=z;
            }
        return;
        }
    i=0;
    ls=0;
    notstarted=1;
    start=-1;
    fstep=1.;
    stepval=0;
    while (1)
        {
        for (;buf[i]==' ' || buf[i]==',' || buf[i]=='\t';i++);
        if (buf[i]=='/')
            {
            i++;
            stepval=1;
            }
        else
            stepval=0;
        if (buf[i]==EOS)
            break;
        for (j=0;buf[i]!=' ' && buf[i]!=',' && buf[i]!='\t' && buf[i]!=EOS;i++)
            buf2[j++]=buf[i];
        if (!j)
            if (buf[i]==EOS)
                break;
            else
                continue;
        buf2[j]=EOS;
        val=atof(buf2);
        if (!stepval && !ls)
            {
            if (!notstarted)
                add_boundaries(start,start,0.);
            start=val;
            notstarted=0;
            continue;
            }
        if (!stepval && ls)
            {
            add_boundaries(start,val,fstep);
            notstarted=1;
            ls=0;
            continue;
            }
        if (stepval && !ls)
            {
            if (notstarted)
                start=d->min;
            fstep=val;
            ls=1;
            continue;
            }
        if (stepval && ls)
            {
            add_boundaries(start,fstep<0 ? d->min : d->max,fstep);
            start=d->min;
            notstarted=0;
            fstep=val;
            continue;
            }
        }
    if (!ls && !notstarted)
        add_boundaries(start,start,0.);
    else
        if (ls)
            add_boundaries(start,fstep<0. ? d->min : d->max,fstep);
    }


static void add_boundaries(float start,float end,float step)

    {
    int     i,n;
    float   z;

    if (nlayers>=MAXLAYERS)
        return;
    if (step==0)
        {
        nlayers++;
        zboundary[nlayers]=start;
        return;
        }
    else
        {
        n=(end-start)/step+.001;
        if (!n)
            {
            nlayers++;
            zboundary[nlayers]=start;
            return;
            }
        else
            for (i=0;i<=n;i++)
                {
                z=start+i*step;
                if (fabs(z/step)<1e-4)
                    z=0;
                nlayers++;
                zboundary[nlayers]=z;
                if (nlayers>=MAXLAYERS)
                    return;
                }
        }
    }


static void plot3d(float far *far*sd,float far *far*in,DATAPTR dp,int grids)

    {
    int     key,x,y,not_clear,grey_scale;
    char    buf[100];
    char    prompt[100];

    if (graphics_on()!=NO_ERROR)
        return;
    grey_scale=0;
    colors3d(grey_scale);
    setup3d(viewtype,dp);
    setupaxes(dp,grids);
    setfillstyle(SOLID_FILL,13);
    diagonal=1;
    if (xscale>25 || yscale>25)
        for (y=yoff+yscale-1;y>=yoff;y--)
            {
            for (x=xoff;x<xoff+xscale;x++)
                tile3d(x,y,sd[x][y],x,y+1,sd[x][y+1],x+1,y+1,sd[x+1][y+1],
                       x+1,y,sd[x+1][y]);
            if (bbioskey(1))
                break;
            }
    else
        for (y=yoff+yscale-1;y>=yoff;y--)
            {
            if (bbioskey(1))
                break;
            for (x=xoff;x<xoff+xscale;x++)
                triangle3d(x+.5,y+.5,in[x][y],x,y+1,sd[x][y+1],x+1,y+1,sd[x+1][y+1]);
            if (bbioskey(1))
                break;
            for (x=xoff;x<xoff+xscale;x++)
                {
                triangle3d(x+1,y,sd[x+1][y],x+.5,y+.5,in[x][y],x+1,y+1,sd[x+1][y+1]);
                triangle3d(x,y,sd[x][y],x,y+1,sd[x][y+1],x+.5,y+.5,in[x][y]);
                }
            if (bbioskey(1))
                break;
            for (x=xoff;x<xoff+xscale;x++)
                triangle3d(x+1,y,sd[x+1][y],x,y,sd[x][y],x+.5,y+.5,in[x][y]);
            }
    if (bbioskey(1))
        bbioskey(0);
    not_clear=0;
    if (viewtype=='c')
        bbox();
    sprintf(prompt,"Press <p> to print, <t> toggle color, <s> save pic,"
                   " %s<ESC> exit",viewtype=='c' ? "<g> grids, " : "");
    ps_update(0,14,prompt,NULL);
    if (viewtype=='c')
        {
        grids=0;
        putgrid(0);
        }
    while (1)
        {
        key=(bbioskey(0)&0xff);
        key=tolower(key);
        if (not_clear)
            {
            not_clear=0;
            ps_clr_update();
            }
        if (key=='t')
            {
            grey_scale=!grey_scale;
            colors3d(grey_scale);
            continue;
            }
        if (key=='p')
            {
            int     append,eject;

            if (nextname(14,buf,&append,&eject))
                {
                ps_3d(buf,dp,xoff,xscale,yoff,yscale,zmin,zscale,viewtype,grids,
                      !grey_scale,zboundary,nlayers,startlayer,append,eject);
                not_clear=1;
                }
            else
                ps_update(0,14,prompt,NULL);
            continue;
            }
        if (viewtype=='c' && key=='g')
            {
            if (grids && erase_grids())
                grids=0;
            else
                {
                putgrid(1);
                grids=1;
                }
            continue;
            }
        if (key=='s')
            {
            ps_update(0,15,vgasave(),NULL);
            not_clear=1;
            continue;
            }
        break;
        }
    free_grids();
    graphics_off();
    }


static void colors3d(int grey_scale)

    {
    int     i;

    if (viewtype=='c')
        {
        contour_colors(!grey_scale);
        return;
        }
    if (grey_scale)
        for (i=0;i<16;i++)
            rgbset(palette3Dg[i][0],palette3Dg[i][1],palette3Dg[i][2],i);
    else
        for (i=0;i<16;i++)
            rgbset(palette3Dc[i][0],palette3Dc[i][1],palette3Dc[i][2],i);
    }


static void triangle3d(float x1,float y1,float z1,float x2,float y2,
                       float z2,float x3,float y3,float z3)

    {
    int p[14];
    int top;
    int i,j,k;
    float   xa[7],ya[7],za[7];
    float   zlo,zhi,zmax;

    zmax=zmin+zscale;
    if ((z1<zmin && z2<zmin && z3<zmin) || (z1>=zmax && z2>=zmax && z3>=zmax))
        return;
    xa[0]=x3d(x1,y1); ya[0]=y3d(x1,y1,z1);
    xa[1]=x3d(x2,y2); ya[1]=y3d(x2,y2,z2);
    xa[2]=x3d(x3,y3); ya[2]=y3d(x3,y3,z3);
    top=((xa[0]-xa[1])*(ya[2]-ya[1])-(xa[2]-xa[1])*(ya[0]-ya[1])<0);
    for (i=0;i<nlayers;i++)
        {
        zlo=zboundary[i];
        zhi=zboundary[i+1];
        j=0;
        if ((z1<zlo && z2<zlo && z3<zlo) || (z1>=zhi && z2>=zhi && z3>=zhi))
            continue;
        clip_line(x1,y1,z1,x2,y2,z2,zlo,zhi,xa,ya,za,&j);
        clip_line(x2,y2,z2,x3,y3,z3,zlo,zhi,xa,ya,za,&j);
        clip_line(x3,y3,z3,x1,y1,z1,zlo,zhi,xa,ya,za,&j);
        for (k=0;k<j;k++)
            {
            p[k<<1] = x3d(xa[k],ya[k]);
            p[(k<<1)+1] = y3d(xa[k],ya[k],za[k]);
            }
        if (top)
            top_pen(i);
        else
            bot_pen(i);
        if (viewtype!='c')
            wire_pen(i);
        fillpoly(j,(int far *)p);
        }
    }


void clip_line(float x1,float y1,float z1,float x2,float y2,float z2,
               float zlo,float zhi,float *xa,float *ya,float *za,
               int *j)

    {
    float   a;

    if ((z1<zlo && z2<zlo) || (z1>=zhi && z2>=zhi))
        return;
    if (z1<zlo)
        {
        a=(zlo-z1)/(z2-z1);
        xa[(*j)]=x1+a*(x2-x1);
        ya[(*j)]=y1+a*(y2-y1);
        za[(*j)]=z1+a*(z2-z1);
        (*j) = (*j)+1;
        a=(z2>=zhi) ? (z2-zhi)/(z2-z1) : 0;
        xa[(*j)]=x2+a*(x1-x2);
        ya[(*j)]=y2+a*(y1-y2);
        za[(*j)]=z2+a*(z1-z2);
        (*j) = (*j)+1;
        return;
        }
    if (z1>=zhi)
        {
        a=(z1-zhi)/(z1-z2);
        xa[(*j)]=x1+a*(x2-x1);
        ya[(*j)]=y1+a*(y2-y1);
        za[(*j)]=z1+a*(z2-z1);
        (*j) = (*j)+1;
        a=(z2<zlo) ? (zlo-z2)/(z1-z2) : 0;
        xa[(*j)]=x2+a*(x1-x2);
        ya[(*j)]=y2+a*(y1-y2);
        za[(*j)]=z2+a*(z1-z2);
        (*j) = (*j)+1;
        return;
        }
    xa[(*j)]=x1;
    ya[(*j)]=y1;
    za[(*j)]=z1;
    (*j) = (*j) + 1;
    if (z2<zlo)
        {
        a=(zlo-z2)/(z1-z2);
        xa[(*j)]=x2+a*(x1-x2);
        ya[(*j)]=y2+a*(y1-y2);
        za[(*j)]=z2+a*(z1-z2);
        (*j) = (*j)+1;
        return;
        }
    if (z2>=zhi)
        {
        a=(z2-zhi)/(z2-z1);
        xa[(*j)]=x2+a*(x1-x2);
        ya[(*j)]=y2+a*(y1-y2);
        za[(*j)]=z2+a*(z1-z2);
        (*j) = (*j)+1;
        }
    }


static void tile3d(float x1,float y1,float z1,float x2,float y2,float z2,
                   float x3,float y3,float z3,float x4,float y4,float z4)

    {
/* Order is important here */
    triangle3d(x1,y1,z1,x2,y2,z2,x3,y3,z3);
    triangle3d(x1,y1,z1,x3,y3,z3,x4,y4,z4);
    }


static void setup3d(int viewtype,DATAPTR dp)

    {
    float   xmax,ymax;

    xmax=getmaxx();
    ymax=getmaxy();
    if (viewtype=='a')
        {
        gyx=0.;
        gyy=-ymax*(1-TM-BM-MAINY)/(float)yscale;
        gyz=-ymax*MAINY/zscale;
        gy0=ymax*(TM+MAINY)-(yoff+yscale)*gyy-zmin*gyz;
        gxx=xmax*MAINX/(float)xscale;
        gxy=xmax*(1-RM-LM-MAINX)/(float)yscale;
        gx0=xmax*LM-xoff*gxx-yoff*gxy;
        }
    else if (viewtype=='b')
        {
        gxx=xmax*(1-RM-LM)/xscale/2.;
        gxy=xmax*(1-RM-LM)/yscale/2.;
        gx0=xmax*LM-xoff*gxx-yoff*gxy;
        gyx=ymax*(1-TM-BM)*VFRACB/2./xscale;
        gyy=-ymax*(1-TM-BM)*VFRACB/2./yscale;
        gyz=-ymax*(1-TM-BM)*(1.-VFRACB)/zscale;
        gy0=ymax*(1-BM)-ymax*(1-TM-BM)*VFRACB/2.-xoff*gyx-yoff*gyy-zmin*gyz;
        }
    else
        {
        gxx=xmax*(1-RM2D-LM2D)/xscale;
        gxy=0.;
        gx0=xmax*(LM2D)-xoff*gxx;
        gyx=0.;
        gyy=-ymax*(1-TM2D-BM2D)/yscale;
        gyz=0.;
        gy0=ymax*(1-BM2D)-yoff*gyy;

        setgriddim(dp->xmin+xoff*(dp->xmax-dp->xmin)/(dp->xsize-1.),
                   dp->ymin+yoff*(dp->ymax-dp->ymin)/(dp->ysize-1.),
                   dp->xmin+(xoff+xscale)*(dp->xmax-dp->xmin)/(dp->xsize-1.),
                   dp->ymin+(yoff+yscale)*(dp->ymax-dp->ymin)/(dp->ysize-1.),
                   xmax*LM2D,ymax*(1-BM2D),
                   xmax*(1-RM2D),ymax*TM2D,
                   "%g","%g");
/*
        setgriddim(xoff,yoff,xoff+xscale,yoff+yscale,
                   xmax*LM2D,ymax*(1-BM2D),
                   xmax*(1-RM2D),ymax*TM2D,
                   "%g","%g");
*/
        }
    }


static int x3d(float x,float y)

    {
    return(x*gxx+y*gxy+gx0);
    }


static int y3d(float x,float y,float z)

    {
    return(x*gyx+y*gyy+z*gyz+gy0);
    }


static void line3d(float x1,float y1,float z1,float x2,float y2,float z2)

    {
    line(x3d(x1,y1),y3d(x1,y1,z1),x3d(x2,y2),y3d(x2,y2,z2));
    }


static void setupaxes(DATAPTR d,int grids)

    {
    char    buf[100];
    int     i,j,i1,i2,modval;
    float   offset,scale,val;
    float   min,max,pval;
    float   z,ztext,x;
    float   dx,dy;

    setcolor(AXES);
    if (viewtype=='c')
        {
        z_legend(d);
        put_titles(d,320,10,x3d(xoff+xscale/2.,0.),y3d(0.,yoff,0.)+7,
                            x3d(xoff,0.)-10,y3d(0.,yoff+yscale/2.,0.));
        return;
        }


    /* X-AXIS */

    if (d->xsize<=1)
        {
        min=d->xmin;
        max=d->xmax;
        }
    else
        {
        min=d->xmin+(float)xoff*(d->xmax-d->xmin)/(d->xsize-1.);
        max=d->xmin+(float)(xoff+xscale)*(d->xmax-d->xmin)/(d->xsize-1.);
        }
    getticks(min,max,&offset,&scale,&i1,&i2,&modval,20);
    if (viewtype=='a')
        settextjustify(CENTER_TEXT,TOP_TEXT);
    else
        settextjustify(RIGHT_TEXT,TOP_TEXT);
    for (i=i1;i<=i2;i++)
        {
        pval=offset+i*scale;
        if (fabs(pval/scale)<1e-5)
            pval=0;
        val=xindex(d,pval);
        if ((i%modval)==0)
            dy=yscale/50.;
        else
            dy=yscale/80.;
        if (nlayers<=1)
            setcolor(AXESBRIGHT);
        else
            top_pen(0);
        line3d(val,yoff,zmin,val,yoff-dy,zmin);
        setcolor(AXES);
        dy=dy*1.5;
        if (i%modval)
            continue;
        if (grids)
            {
            for (j=0;j<nlayers;j++)
                {
                if (nlayers>1)
                    bot_pen(j);
                line3d(val,yoff+yscale,zboundary[j],
                       val,yoff+yscale,zboundary[j+1]);
                }
            if (nlayers>1)
                bot_pen(0);
            line3d(val,yoff,zmin,val,yoff+yscale,zmin);
            }
        sprintf(buf,"%g",pval);
        setcolor(AXESBRIGHT);
        outtextxy(x3d(val,yoff-dy),y3d(val,yoff-dy,zmin)+2,buf);
        setcolor(AXES);
        }

    /* Y-AXIS */

    if (d->ysize<=1)
        {
        min=d->ymin;
        max=d->ymax;
        }
    else
        {
        min=d->ymin+(float)yoff*(d->ymax-d->ymin)/(d->ysize-1.);
        max=d->ymin+(float)(yoff+yscale)*(d->ymax-d->ymin)/(d->ysize-1.);
        }
    getticks(min,max,&offset,&scale,&i1,&i2,&modval,20);
    settextjustify(LEFT_TEXT,TOP_TEXT);
    x=xoff+xscale;
    z=zmin;
    for (i=i1+1;i<=i2;i++)
        {
        pval=offset+i*scale;
        if (fabs(pval/scale)<1e-5)
            pval=0;
        val=yindex(d,pval);
        if ((i%modval)==0)
            dx=-xscale/80.;
        else
            dx=-xscale/100.;
        if (nlayers<=1)
            setcolor(AXESBRIGHT);
        else
            top_pen(0);
        line3d(x-dx,val,z,x,val,z);
        dx=dx*1.5;
        setcolor(AXES);
        if (i%modval)
            continue;
        if (grids)
            {
            for (j=0;j<nlayers;j++)
                {
                if (nlayers>1)
                    bot_pen(j);
                line3d(xoff,val,zboundary[j],xoff,val,zboundary[j+1]);
                }
            if (nlayers>1)
                bot_pen(0);
            line3d(xoff,val,zmin,xoff+xscale,val,zmin);
            }
        sprintf(buf,"%g",pval);
        setcolor(AXESBRIGHT);
        outtextxy(x3d(x-dx,val),y3d(x-dx,val,z),buf);
        setcolor(AXES);
        }

    /* Z-AXIS */

    settextjustify(RIGHT_TEXT,CENTER_TEXT);
    dy=yscale/50.;
    if (nlayers>1)
        {
        ztext=zmin-zscale;
        for (i=0;i<nlayers;i++)
            {
            z=zboundary[i];
            if (grids)
                {
                bot_pen(i);
                line3d(xoff,yoff+yscale,z,xoff,yoff,z);
                line3d(xoff,yoff+yscale,z,xoff+xscale,yoff+yscale,z);
                }
            top_pen(i);
            line3d(xoff,yoff-dy,z,xoff,yoff,z);
            if ((z-ztext)/zscale > 0.03)
                {
                sprintf(buf,"%g",z);
                outtextxy(x3d(xoff,yoff-dy*1.5)-2,
                          y3d(xoff,yoff-dy*1.5,z),buf);
                ztext=z;
                }
            three_verts(grids,z,zboundary[i+1]);
            }
        if ((zmin+zscale-ztext)/zscale > 0.03)
            {
            top_pen(nlayers-1);
            line3d(xoff,yoff-dy,zmin+zscale,xoff,yoff,zmin+zscale);
            sprintf(buf,"%g",zmin+zscale);
            outtextxy(x3d(xoff,yoff-dy*1.5)-2,
                      y3d(xoff,yoff-dy*1.5,zmin+zscale),buf);
            }
        }
    else
        {
        getticks(zmin,zmin+zscale,&offset,&scale,&i1,&i2,&modval,20);
        for (i=i1+1;i<=i2;i++)
            {
            z=offset+i*scale;
            if (fabs(z/scale)<1e-5)
                z=0;
            if ((i%modval)==0)
                dy=yscale/80.;
            else
                dy=yscale/100.;
            setcolor(AXESBRIGHT);
            line3d(xoff,yoff,z,xoff,yoff-dy,z);
            setcolor(AXES);
            if (i%modval)
                continue;
            dy=dy*1.5;
            if (grids)
                {
            line3d(xoff,yoff+yscale,z,xoff,yoff,z);
            line3d(xoff,yoff+yscale,z,xoff+xscale,yoff+yscale,z);
                }
            sprintf(buf,"%g",z);
            setcolor(AXESBRIGHT);
            outtextxy(x3d(xoff,yoff-dy)-2,y3d(xoff,yoff-dy,z),buf);
            setcolor(AXES);
            }
        }

    /* Draw 3-D outline box */

    setcolor(AXESBRIGHT);
    if (nlayers>1)
        top_pen(0);
    line3d(xoff+xscale,yoff,zmin,xoff,yoff,zmin);
    line3d(xoff+xscale,yoff+yscale,zmin,xoff+xscale,yoff,zmin);
    if (nlayers<=1)
        line3d(xoff,yoff,zmin,xoff,yoff,zmin+zscale);
    if (grids)
        {
        if (nlayers>1)
            top_pen(0);
        line3d(xoff,yoff+yscale,zmin,xoff+xscale,yoff+yscale,zmin);
        line3d(xoff,yoff,zmin,xoff,yoff+yscale,zmin);
        if (nlayers<=1)
            {
            setcolor(AXESBRIGHT);
            line3d(xoff,yoff+yscale,zmin,xoff,yoff+yscale,zmin+zscale);
            line3d(xoff+xscale,yoff+yscale,zmin,
                   xoff+xscale,yoff+yscale,zmin+zscale);
            }
        if (nlayers>1)
            top_pen(nlayers-1);
        line3d(xoff,yoff,zmin+zscale,xoff,yoff+yscale,zmin+zscale);
        line3d(xoff,yoff+yscale,zmin+zscale,
               xoff+xscale,yoff+yscale,zmin+zscale);
        }

    /* Put titles */
    setcolor(AXESBRIGHT);
    settextjustify(CENTER_TEXT,TOP_TEXT);
    outtextxy(320,10,d->gtitle);
    settextstyle(0,VERT_DIR,0);
    settextjustify(RIGHT_TEXT,CENTER_TEXT);
    outtextxy(x3d(xoff,yoff)-60,y3d(xoff,yoff,zmin+zscale/2.),d->ztitle);
    settextstyle(0,HORIZ_DIR,0);

    if (viewtype=='a')
        {
        settextjustify(CENTER_TEXT,TOP_TEXT);
        outtextxy(x3d(xoff+xscale/2.,yoff),y3d(xoff+xscale/2.,yoff,zmin)+14,
                  d->xtitle);
        settextjustify(LEFT_TEXT,TOP_TEXT);
        outtextxy(x3d(xoff+xscale,yoff+yscale/2.)+30,
                  y3d(xoff+xscale,yoff+yscale/2.,zmin)+10,
                  d->ytitle);
        }
    else
        {
        settextjustify(RIGHT_TEXT,TOP_TEXT);
        outtextxy(x3d(xoff+xscale/2.,yoff)-30,y3d(xoff+xscale/2.,yoff,zmin)+10,
                  d->xtitle);
        settextjustify(LEFT_TEXT,TOP_TEXT);
        outtextxy(x3d(xoff+xscale,yoff+yscale/2.)+30,
                  y3d(xoff+xscale,yoff+yscale/2.,zmin)+10,
                  d->ytitle);
        }
    }


static void z_legend(DATAPTR d)

    {
    int     i,ytext,x1,x2,yt,yb,y0,y1,xmax,ymax;
    float   a0,a1;
    char    buf[40];

    xmax=getmaxx();
    ymax=getmaxy();
    ytext=ymax+1000;
    x1=LEGMAR*xmax;
    x2=x1+20;
    yb=y3d(xoff,yoff,0);
    yt=y3d(xoff,yoff+yscale,0);
    settextjustify(LEFT_TEXT,CENTER_TEXT);
    for (i=0;i<nlayers;i++)
        {
        a0=(zboundary[i]-zmin)/zscale;
        a1=(zboundary[i+1]-zmin)/zscale;
        y0=y3d(xoff,yoff+yscale*a0,0);
        y1=y3d(xoff,yoff+yscale*a1,0);
        top_pen(i);
        bar(x1,y0,x2,y1);
        setcolor(AXESBRIGHT);
        line(x2,y0,x2+4,y0);
        if (ytext-y0 > 8)
            {
            sprintf(buf,"%g",zboundary[i]);
            outtextxy(x2+6,y0,buf);
            ytext=y0;
            }
        }
    setcolor(AXESBRIGHT);
    line(x2,yt,x2+4,yt);
    if (ytext-yt>10)
        {
        sprintf(buf,"%g",zboundary[nlayers]);
        outtextxy(x2+6,yt,buf);
        }
    rectangle(x1,yt,x2,yb);
    settextstyle(0,VERT_DIR,0);
    settextjustify(RIGHT_TEXT,CENTER_TEXT);
    outtextxy((int)(getmaxx()*LEGMAR)-3,
              y3d(xoff,yoff+yscale/2.,0.),d->ztitle);
    settextstyle(0,HORIZ_DIR,0);
    }


static void three_verts(int grids,float z0,float z1)

    {
    line3d(xoff,yoff,z0,xoff,yoff,z1);
    if (grids)
        {
        line3d(xoff,yoff+yscale,z0,xoff,yoff+yscale,z1);
        line3d(xoff+xscale,yoff+yscale,z0,xoff+xscale,yoff+yscale,z1);
        }
    }


static void top_pen(int layer)

    {
    int     c;

    if (viewtype=='c')
        c=9-(layer%8);
    else
        c=layerpen[(layer+startlayer+NLAYERCOLORS)%NLAYERCOLORS][0];
    setcolor(c);
    setfillstyle(SOLID_FILL,c);
    }


static void bot_pen(int layer)

    {
    int     c;

    c=layerpen[(layer+startlayer+NLAYERCOLORS)%NLAYERCOLORS][1];
    setcolor(c);
    setfillstyle(SOLID_FILL,c);
    }


static void wire_pen(int layer)

    {
    setcolor(layerpen[(layer+startlayer+NLAYERCOLORS)%NLAYERCOLORS][2]);
    }
