/*
**
** AREA.C      Calculates the area/volume inside a contour.
**
** Will Menninger, 9/89, programmed in Turbo C 2.0
**
** Modified 6-1-92 for any size rectangular grid.
**
*/

#include "contour.h"

#define CRDCPY(a,b)     t1.x[b]=t0.x[a];t1.y[b]=t0.y[a];t1.z[b]=t0.z[a];
#define CRDINT(a,b,c)   t1.x[c]=t0.x[a]+(t0.x[b]-t0.x[a])*(z-t0.z[a])/      \
                                                         (t0.z[b]-t0.z[a]); \
                        t1.y[c]=t0.y[a]+(t0.y[b]-t0.y[a])*(z-t0.z[a])/      \
                                                         (t0.z[b]-t0.z[a]); \
                        t1.z[c]=z;
#define SIDELEN(a,b)    sqrt((t->x[b]-t->x[a])*(t->x[b]-t->x[a])+           \
                             (t->y[b]-t->y[a])*(t->y[b]-t->y[a]))
#define MAX3I(a,b,c)    a>b ? (a>c ? 0 : 2) : (b>c ? 1 : 2)


/*
**  TRIANGLE COORDINATES ARE DONE AS FOLLOWS:
**
**        ___________________________________
**       |\      /|\      /|\      /|\      /|
**       | \ 10 / | \ 30 / | \ 50 / | \ 70 / |
**       |  \  /  |  \  /  |  \  /  |  \  /  |
**       |00 \/   |20 \/   |40 \/   |60 \/   |
** XY =  |   /\ 11|   /\ 31|   /\ 51|   /\ 71|
**       |  /  \  |  /  \  |  /  \  |  /  \  |
**       | / 01 \ | / 21 \ | / 41 \ | / 61 \ |
**       |/______\|/______\|/______\|/______\|    . . .
**       |\      /|\      /|\      /|\      /|
**       | \ 12 / | \ 32 / | \ 52 / | \ 72 / |
**       |  \  /  |  \  /  |  \  /  |  \  /  |
**       |02 \/   |22 \/   |42 \/   |62 \/   |
**       |   /\ 13|   /\ 33|   /\ 53|   /\ 73|
**       |  /  \  |  /  \  |  /  \  |  /  \  |
**       | / 03 \ | / 23 \ | / 43 \ | / 63 \ |
**       |/______\|/______\|/______\|/______\|
**
**
**                         .
**
**                         .
**
**                         .
**
**
*/

typedef struct AP   {
                    float       area;
                    float       volume;
                    float       vcx;
                    float       vcy;
                    float       acx;
                    float       acy;
                    float       zmax;
                    float       zmx;
                    float       zmy;
                    float       xmin;
                    float       xmax;
                    float       ymin;
                    float       ymax;
                    int         ntris;
                    struct AP  *next;
                    } AREA, *AREAPTR;

typedef struct  {
                float   x[4],y[4],z[4];
                } TRI, *TRIPTR;

static unsigned char    far*done;
static unsigned int         donesize;
static int             huge*stack;
static long                 stackspace;
static long                 stackptr;
static int                  xb1,xb2,yb1,yb2;
static int                  xsize;
static int                  ysize;
static int                  xs2;
static int                  ys2;
static float        far*far*sd;
static float        far*far*in;
static DATAPTR              dp;

static AREAPTR find_areas     (float z);
static void    free_areas     (AREAPTR h);
static void    intersect      (float xa1,float ya1,float xa2,float ya2,
                               float xb1,float yb1,float xb2,float yb2,
                               float *x,float *y);
static void    vol_median     (float c,float d,float g,float *b);
static BOOLEAN correct_height (int x,int y,float z);
static void    calc_area      (int x,int y,float z,AREAPTR a);
static void    pushneighbor   (int x,int y,int nt);
static void    cleararea      (AREAPTR a);
static BOOLEAN poptri         (int *x,int *y);
static STATUS  pushtri        (int x,int y);
static void    clearstack     (void);
static BOOLEAN tridone        (int x,int y);
static void    settridone     (int x,int y);
static void    setallnotdone  (void);
static void    do_tri         (AREAPTR a,TRIPTR t,float z);


void init_avars(DATAPTR x)

    {
    if (!loaded(1))
        return;
    dp=x;
    sd=x->shotdata;
    in=x->interp;
    xsize=x->xsize;
    ysize=x->ysize;
    stackspace=x->stacksize;
    stack=x->stack;
    xs2=(xsize-1)<<1;
    ys2=(ysize-1)<<1;
    donesize=(unsigned int)x->donesize;
    done=x->done;
    }


void area_stuff(void)

    {
    char s[50];
    float z;
    AREAPTR h,p;
    int     i;
    float   xf1,yf1,xf2,yf2;
    float   ascale;

    if (!loaded(TRUE))
        return;
    while (1)
        {
        if (bgetinput("Enter bounding box (x1,y1,x2,y2):  ",s,40)==ESCAPE)
            return;
        normal("\r\n\r\n");
        if (sscanf(s,"%f,%f,%f,%f",&xf1,&yf1,&xf2,&yf2)<4)
            {
            xb1=0;
            yb1=0;
            xb2=xs2;
            yb2=ys2;
            break;
            }
        xb1=xindex(dp,xf1);
        yb1=yindex(dp,yf1);
        xb2=xindex(dp,xf2)+.99;
        yb2=yindex(dp,yf2)+.99;
        xb1*=2;
        yb1*=2;
        xb2*=2;
        yb2*=2;

        if (xb1>xb2)
            {
            i=xb1; xb1=xb2; xb2=i;
            }
        if (yb1>yb2)
            {
            i=yb1; yb1=yb2; yb2=i;
            }
        if (xb1<0)
            xb1=0;
        if (xb2>xs2)
            xb2=xs2;
        if (yb1<0)
            yb1=0;
        if (yb2>ys2)
            yb2=ys2;
        if (xb1>xs2 || xb2<0 || yb1>ys2 || yb2<0)
            {
            error("Bounding box is illegal.");
            normal("\r\n\r\n");
            }
        else
            break;
        }
    while (1)
        {
        if (bgetinput("Enter contour height:  ",s,20)==ESCAPE)
            return;
        normal("\r\n\r\n");
        z=atof(s);
        break;
        }
    h=find_areas(z);
    p=h;
    if (p==NULL)
        cputs("No contours found.\r\n");
    else
        {
        ascale=(dp->xmax-dp->xmin)*(dp->ymax-dp->ymin)/((float)(dp->xsize-1.)*(dp->ysize-1.));
        for (i=1;p!=NULL;p=p->next,i++)
            {
            p->acx=dp->xmin+p->acx*(dp->xmax-dp->xmin)/(dp->xsize-1.);
            p->acy=dp->ymin+p->acy*(dp->ymax-dp->ymin)/(dp->ysize-1.);
            p->vcx=dp->xmin+p->vcx*(dp->xmax-dp->xmin)/(dp->xsize-1.);
            p->vcy=dp->ymin+p->vcy*(dp->ymax-dp->ymin)/(dp->ysize-1.);
            p->area*=ascale;
            p->volume*=ascale;
            p->xmin=dp->xmin+p->xmin*(dp->xmax-dp->xmin)/(dp->xsize-1.);
            p->xmax=dp->xmin+p->xmax*(dp->xmax-dp->xmin)/(dp->xsize-1.);
            p->ymin=dp->ymin+p->ymin*(dp->ymax-dp->ymin)/(dp->ysize-1.);
            p->ymax=dp->ymin+p->ymax*(dp->ymax-dp->ymin)/(dp->ysize-1.);
            cprintf("Contour #%d:  (%g,%g) (%g,%g), area=%g vol=%g\r\n",
                    i,p->acx,p->acy,p->vcx,p->vcy,p->area,p->volume);
            cprintf("   Bounding box:  (%g,%g)-(%g,%g)\r\n",
                    p->xmin,p->ymin,p->xmax,p->ymax);
            }
        }
    free_areas(h);
    wait_enter();
    }


static AREAPTR find_areas(float z)

    {
    int     x,y,x0,y0;
    AREAPTR h,p;

    h=NULL;
    p=NULL;
    setallnotdone();
    clearstack();

    for (y=yb1;y<yb2;y++)
        for (x=xb1;x<xb2;x++)
            {
            if (tridone(x,y))
                continue;
            if (correct_height(x,y,z))
                {
                cprintf("Valid region found at cell (%4d,%4d)\r\n",x,y);
                if (h==NULL)
                    {
                    h=calloc(1,sizeof(AREA));
                    if (h==NULL)
                        {
                        error("Out of memory!");
                        normal("\r\n");
                        return(NULL);
                        }
                    p=h;
                    }
                else
                    {
                    h->next=calloc(1,sizeof(AREA));
                    if (h->next==NULL)
                        {
                        error("Out of memory!");
                        normal("\r\n");
                        return(p);
                        }
                    h=h->next;
                    }
                cleararea(h);
                calc_area(x,y,z,h);
                settridone(x,y);
                while (poptri(&x0,&y0))
                    calc_area(x0,y0,z,h);
                }
            else
                settridone(x,y);
            }
    return(p);
    }


static void free_areas(AREAPTR h)

    {
    AREAPTR p;

    while (h!=NULL)
        {
        p=h->next;
        free(h);
        h=p;
        }
    }


static BOOLEAN correct_height(int x,int y,float z)

    {
    int     ym,xm,x2,y2;

    xm=x&1;
    ym=y&1;
    x2=x>>1;
    y2=y>>1;
    return(z<=in[x2][y2] || z<=sd[x2+ym][y2+ym] || z<=sd[x2+xm][y2+1-xm]);
    }


static void calc_area(int x,int y,float z,AREAPTR a)

    {
    int     ym,xm;
    TRI     t0;
    TRI     t1;

    xm=x&1;
    ym=y&1;

    t0.x[0]=x/2+.5;
    t0.x[1]=x/2+ym;
    t0.x[2]=x/2+xm;

    t0.y[0]=y/2+.5;
    t0.y[1]=y/2+ym;
    t0.y[2]=y/2+1-xm;

    t0.z[0]=in[x>>1][y>>1];
    t0.z[1]=sd[(x>>1)+ym][(y>>1)+ym];
    t0.z[2]=sd[(x>>1)+xm][(y>>1)+1-xm];

    t1.z[3]=-1;

    if (z<=t0.z[0])
        {
        pushneighbor(x,y,0);
        pushneighbor(x,y,1);
        CRDCPY(0,0);
        if (z<=t0.z[1])
            {
            pushneighbor(x,y,2);
            CRDCPY(1,1);
            if (z<=t0.z[2])
                {
                CRDCPY(2,2);
                }
            else
                {
                CRDINT(0,2,3);
                CRDINT(1,2,2);
                }
            }
        else
            if (z<=t0.z[2])
                {
                pushneighbor(x,y,2);
                CRDCPY(2,1);
                CRDINT(1,2,2);
                CRDINT(0,1,3);
                }
            else
                {
                CRDINT(0,1,1);
                CRDINT(0,2,2);
                }
        }
    else
        {
        pushneighbor(x,y,2);
        if (z<=t0.z[1])
            {
            pushneighbor(x,y,0);
            CRDCPY(1,0);
            if (z<=t0.z[2])
                {
                pushneighbor(x,y,1);
                CRDCPY(2,1);
                CRDINT(0,2,2);
                CRDINT(0,1,3);
                }
            else
                {
                CRDINT(1,0,1);
                CRDINT(1,2,2);
                }
            }
        else
            {
            pushneighbor(x,y,1);
            CRDCPY(2,0);
            CRDINT(1,2,1);
            CRDINT(0,2,2);
            }
        }
    do_tri(a,&t1,z);
    if (t1.z[3]>0)
        {
        t1.x[1]=t1.x[3];
        t1.y[1]=t1.y[3];
        t1.z[1]=t1.z[3];
        do_tri(a,&t1,z);
        }
    }


static void do_tri(AREAPTR a,TRIPTR t,float z)

    {
    int     i;
    float   d,e,f,phi,beta;
    float   x1,x2,y1,y2,cx,cy,da,dv;

    a->ntris++;
    for (i=0;i<3;i++)
        {
        if (t->x[i]<a->xmin)
            a->xmin=t->x[i];
        if (t->x[i]>a->xmax)
            a->xmax=t->x[i];
        if (t->y[i]<a->ymin)
            a->ymin=t->y[i];
        if (t->y[i]>a->ymax)
            a->ymax=t->y[i];
        }
    i=MAX3I(t->z[0],t->z[1],t->z[2]);
    if (t->z[i]>a->zmax)
        {
        a->zmax=t->z[i];
        a->zmx=t->x[i];
        a->zmy=t->y[i];
        }
    d=SIDELEN(0,1);
    e=SIDELEN(1,2);
    f=SIDELEN(2,0);
    if (d==0 || e==0 || f==0)
        return;
    beta=(d*d+e*e-f*f)/(2*d*e);
    if (beta>1)
        {
        error("Funny triangle:  beta > 1!");
        normal("\r\n\r\n");
        return;
        }
    beta=sqrt(1-beta*beta);

    x1=(t->x[0]+t->x[1])/2;
    y1=(t->y[0]+t->y[1])/2;
    x2=(t->x[2]+t->x[1])/2;
    y2=(t->y[2]+t->y[1])/2;
    intersect(x1,y1,t->x[2],t->y[2],x2,y2,t->x[0],t->y[0],&cx,&cy);
    da=.5*beta*e*d;
    a->acx=(a->area*a->acx+da*cx)/(a->area+da);
    a->acy=(a->area*a->acy+da*cy)/(a->area+da);
    a->area+=da;

    vol_median(t->z[0],t->z[1],t->z[2],&phi);
    x1=t->x[0]+(t->x[1]-t->x[0])*phi;
    y1=t->y[0]+(t->y[1]-t->y[0])*phi;
    vol_median(t->z[1],t->z[2],t->z[0],&phi);
    x2=t->x[1]+(t->x[2]-t->x[1])*phi;
    y2=t->y[1]+(t->y[2]-t->y[1])*phi;
    intersect(x1,y1,t->x[2],t->y[2],x2,y2,t->x[0],t->y[0],&cx,&cy);
    dv=beta*e*d*(t->z[0]+t->z[1]+t->z[2]-z*3)/6;
    a->vcx=(a->volume*a->vcx+dv*cx)/(a->volume+dv);
    a->vcy=(a->volume*a->vcy+dv*cy)/(a->volume+dv);
    a->volume+=dv;
    }


static void vol_median(float c,float d,float g,float *b)

    {
    float   alpha;

    alpha=c-d;
    if (fabs(alpha)<.01)
        (*b)=1-.5*(alpha+g+2*d)/(g+2*d);
    else
        {
        alpha=2*c*c+2*c*g+g*g+2*d*g+2*d*d;
        (*b)=(2*c+g-sqrt(alpha))/(c-d);
        }
    }


static void intersect(float xa1,float ya1,float xa2,float ya2,float xb1,
                      float yb1,float xb2,float yb2,float *x,float *y)

    {
    BOOLEAN hsa,hsb;
    float   m1,m2,b1,b2;

    hsa=fabs(ya2-ya1)<fabs(xa2-xa1);
    hsb=fabs(yb2-yb1)<fabs(xb2-xb1);
    if (!hsa)
        {
        m1=(xa2-xa1)/(ya2-ya1);
        b1=-m1*ya1+xa1;
        }
    else
        {
        m1=(ya2-ya1)/(xa2-xa1);
        b1=-m1*xa1+ya1;
        }
    if (!hsb)
        {
        m2=(xb2-xb1)/(yb2-yb1);
        b2=-m2*yb1+xb1;
        }
    else
        {
        m2=(yb2-yb1)/(xb2-xb1);
        b2=-m2*xb1+yb1;
        }
    if (hsa)
        {
        if (hsb)
            (*x)=(b2-b1)/(m1-m2);
        else
            (*x)=(m2*b1+b2)/(1-m2*m1);
        (*y)=m1*(*x)+b1;
        }
    else
        {
        if (hsb)
            (*y)=(m2*b1+b2)/(1-m2*m1);
        else
            (*y)=(b2-b1)/(m1-m2);
        (*x)=m1*(*y)+b1;
        }
    }


static void pushneighbor(int x,int y,int nt)

    {
    int     xm,ym;
    int     xn,yn;

    xm=x&1;
    ym=y&1;
    if (nt==0)
        {
        xn=2*(x>>1)+1-xm;
        yn=y;
        }
    else
        if (nt==1)
            {
            xn=x;
            yn=2*(y>>1)+1-ym;
            }
        else
            {
            xn=x-1+(ym<<1);
            yn=y+1-(xm<<1);
            }
    if (!tridone(xn,yn) && xn>=xb1 && xn<xb2 && yn>=yb1 && yn<yb2)
        {
        pushtri(xn,yn);
        settridone(xn,yn);
        }
    }


static void cleararea(AREAPTR a)

    {
    a->area=0;
    a->volume=0;
    a->vcx=0;
    a->vcy=0;
    a->acx=0;
    a->acy=0;
    a->zmax=0;
    a->zmx=0;
    a->zmy=0;
    a->ntris=0;
    a->xmin=32000;
    a->xmax=-1;
    a->ymin=32000;
    a->ymax=-1;
    a->next=NULL;
    }


static BOOLEAN poptri(int *x,int *y)

    {
    if (stackptr<=0)
        return(FALSE);
    (*y)=stack[stackptr--];
    (*x)=stack[stackptr--];
    return(TRUE);
    }


static STATUS pushtri(int x,int y)

    {
    if (stackptr>=stackspace-2)
        {
        error("Pushtri:  Stack overflow!");
        normal("\r\n");
        return(ERROR);
        }
    stack[++stackptr]=x;
    stack[++stackptr]=y;
    return(NO_ERROR);
    }


static void clearstack(void)

    {
    stackptr=-1;
    }


static BOOLEAN tridone(int x,int y)

    {
    return(done[(x>>3)*ys2+y]&(1<<(x&7)));
    }


static void settridone(int x,int y)

    {
    done[(x>>3)*ys2+y] |= (1<<(x&7));
    }


static void setallnotdone(void)

    {
    unsigned int     i;

    for (i=0;i<donesize;i++)
        done[i]=0;
    }
