/*
**
** AXES.C   Functions that help in deciding how to number an axis
**
** Will Menninger, 11-19-92
**
*/

#include <stdio.h>
#include <math.h>

#include "contour.h"

static float    xmin,xmax,ymin,ymax,zmin,zmax;


void getticks(float x1,float x2,float *offset,float *pertick,
              int *first,int *last,int *modval,int nticks)

    {
    float   s,p,delta,l,firsttick;
    int     tickmod,numticks;

    delta=x2-x1;
    delta/=nticks;
    p=floor(log10(delta));
    s=delta/pow(10.,p);
    if (s<1.5 || s>7.5)
        {
        if (s>7.5)
            p++;
        s=1;       /* Minor ticks every 1, Major every 5 */
        (*modval)=5;
        }
    else if (s<3.5)
        {          /* Minor ticks every 2, Major every 1 */
        s=2;
        (*modval)=5;
        }
    else
        {
        s=5;       /* Minor ticks every 5, Major every 2 */
        (*modval)=4;
        }
    (*pertick)=s*pow(10.,p);
    l=floor((x1+.99*(*pertick))/(*pertick));
    tickmod=(int)l%(*modval);
    firsttick=floor((x1+.99*(*pertick))/(*pertick))*(*pertick);
    numticks=(x2-firsttick+.01*(*pertick))/(*pertick);
    (*offset)=firsttick-(*pertick)*tickmod;
    (*first)=tickmod;
    (*last)=numticks+tickmod;
    }


void init_ranges(DATAPTR d)

    {
    xmin=d->xmin;
    xmax=d->xmax;
    ymin=d->ymin;
    ymax=d->ymax;
    zmin=d->min;
    zmax=d->max;
    }


int get_range(char axis,DATAPTR d,float *min,float *max)

    {
    char    prompt[80];
    char    buf[100];
    int     i,nargs;
    float   a,b;
    float   minmin,maxmax,*defmin,*defmax;

    if (axis=='X')
        {
        minmin=d->xmin;
        maxmax=d->xmax;
        defmin=&xmin;
        defmax=&xmax;
        }
    else if (axis=='Y')
        {
        minmin=d->ymin;
        maxmax=d->ymax;
        defmin=&ymin;
        defmax=&ymax;
        }
    else
        {
        defmin=&zmin;
        defmax=&zmax;
        }
    sprintf(prompt,"Enter %c-axis range [Def=%g,%g]:  ",axis,
            (*defmin),(*defmax));
    while (1)
        {
        if (bgetinput(prompt,buf,20)==ESCAPE)
            return(0);
        normal("\r\n");
        for (i=0;buf[i]==' ' || buf[i]=='\t';i++);
        if (buf[i]=='\0')
            {
            (*min)=(*defmin);
            (*max)=(*defmax);
            }
        else if (buf[i]==',')
            {
            (*min)=(*defmin);
            nargs=sscanf(buf,",%f",&a);
            (*max) = nargs<1 ? (*defmax) : a;
            }
        else
            {
            nargs=sscanf(buf,"%f,%f",&a,&b);
            if (nargs<2)
                nargs=sscanf(buf,"%f %f",&a,&b);
            (*max) = nargs<2 ? (*defmax) : b;
            (*min) = nargs<1 ? (*defmin) : a;
            }
        if ((*min)>=(*max))
            {
            normal("\r\n");
            error("Range must have a non-zero, positive width!");
            normal("\r\n\r\n");
            continue;
            }
        if (axis=='Z')
            {
            (*defmin)=(*min);
            (*defmax)=(*max);
            break;
            }
        if ((*min)>=maxmax || (*max)<=minmin)
            {
            normal("\r\n");
            sprintf(buf,"Range must be within plot bounds (%g to %g).",
                    minmin,maxmax);
            error(buf);
            normal("\r\n\r\n");
            continue;
            }
        if ((*min)<minmin)
            (*min)=minmin;
        if ((*max)>maxmax)
            (*max)=maxmax;
        (*defmin)=(*min);
        (*defmax)=(*max);
        break;
        }
    return(1);
    }
