/*
**
** BATCH.C     Replaces all input routines with routines to check
**             for input from a batch file first.  If the batch
**             file runs out, keyboard input is used.
**
** Will Menninger, 3/95
**
** Part of CONTOUR program.
**
*/


#include "contour.h"

static FILE *bf = NULL;
static char buf[80];


static int get_line(void);
static void clean_line(char *buf);


void setbatch(char *filename)

    {
    bf=fopen(filename,"r");
    if (bf==NULL)
        return;
    }

void killbatch(void)

    {
    if (bf!=NULL)
        {
        fclose(bf);
        bf=NULL;
        }
    }


int bmenu(char *m[])

    {
    int     i;

    if (bf==NULL)
        return(menu(m));
    while (1)
        {
        if (!get_line())
            {
            killbatch();
            return(menu(m));
            }
        i=buf[0]-'0';
        if (i>=1 && i<=9)
            break;
        }
    return(i);
    }


int bbioskey(int arg)

    {
    if (bf==NULL)
        return(bioskey(arg));
    if (arg==1)
        return(0);
    if (!get_line())
        return(bioskey(arg));
    if (buf[0]==EOS)
        return(13);
    return(buf[0]);
    }


int bgetinput(char *prompt,char *mbuf,int maxlen)

    {
    if (bf==NULL || !get_line())
        return(getinput(prompt,mbuf,maxlen));
    if (maxlen<strlen(buf))
        buf[maxlen-1]=EOS;
    strcpy(mbuf,buf);
    return(NO_ERROR);
    }


static int get_line(void)

    {
    int     i;

    while (1)
        {
        if (fgets(buf,79,bf)==NULL)
            return(0);
        clean_line(buf);
        if (buf[0]!=';')
            break;
        }
    for (i=0;buf[i]!=EOS;i++)
        buf[i]=tolower(buf[i]);
    return(1);
    }


static void clean_line(char *buf)

    {
    int     i,j;

    for (i=0;buf[i]!='\n' && buf[i]!='\r' && buf[i]!=EOS;i++);
    for (i--;i>=0 && (buf[i]==' ' || buf[i]=='\t');i--);
    buf[i+1]=EOS;
    for (i=0;buf[i]==' ' || buf[i]=='\t';i++);
    if (i)
        {
        for (j=0;buf[i]!=EOS;j++,i++)
            buf[j]=buf[i];
        buf[j]=EOS;
        }
    }
