/*
**
** CLCALC.C    Finds contour points start to finish.
**
** Will Menninger, programmed in Turbo C 2.0
**
** 1-16-93
**
*/

#include "contour.h"

#define CRDCPY(a,b)     t1.x[b]=t0.x[a];t1.y[b]=t0.y[a];t1.z[b]=t0.z[a];
#define CRDINT(a,b,c)   t1.x[c]=t0.x[a]+(t0.x[b]-t0.x[a])*(z-t0.z[a])/      \
                                                         (t0.z[b]-t0.z[a]); \
                        t1.y[c]=t0.y[a]+(t0.y[b]-t0.y[a])*(z-t0.z[a])/      \
                                                         (t0.z[b]-t0.z[a]); \
                        t1.z[c]=z;
#define SIDELEN(a,b)    sqrt((t->x[b]-t->x[a])*(t->x[b]-t->x[a])+           \
                             (t->y[b]-t->y[a])*(t->y[b]-t->y[a]))
#define MAX3I(a,b,c)    a>b ? (a>c ? 0 : 2) : (b>c ? 1 : 2)

#define OUTOFBOUNDS(x,y,v)  (((x)==xb2 && (v) > 2) || ((y)==yb2 && \
                             ((v)==2 || (v) > 3)))
#define COUNTPOINT(x)   if ((x)==z) return(0); else if ((x)<z) below++; else \
                        above++
#define CKNEIGHBOR(x,y,v)   goodneighbor(x,y,v,xo,yo,vo,xn,yn,vn,&n,nnd)
#define ISWAP(x,y)      {int tt; tt=(x); (x)=(y); (y)=tt;}
#define VSWAP(x1,y1,v1,x2,y2,v2)    ISWAP(x1,x2) ISWAP(y1,y2) ISWAP(v1,v2)
#define VCOPY(x1,y1,v1,x2,y2,v2)    (x1)=(x2); (y1)=(y2); (v1)=(v2)


/*
**  Vertex Coordinates are as follows:
**
**        ________
**       |\      /|
**       | 7    8 |
**       |  \  /  |
**       2    4   |
**       |        |
**       |  /  \  |
**       | 5    6 |
**       1/___3__\|
**
**
**
*/

static int                  xb1,yb1,xb2,yb2;
static long                 vdone,vmax;
static unsigned char    far*done;
static unsigned int         donesize;
static float        far*far*sd;
static float        far*far*in;
static DATAPTR              dp;

static void trace_path(int x,int y,int v,float z);
static int count_points(int x,int y,int v,int x2,int y2,int v2,float z);
static int trace_back(int *x,int *y,int *v,int *x2,int *y2,int *v2,int *kfp,
                      float z);
static BOOLEAN correct_height(int x,int y,int v,float z);
static float xpoint(int x,int y,int v,float z);
static float ypoint(int x,int y,int v,float z);
static BOOLEAN single_point(int x,int y,int v);
static int neighbor(int x,int y,int v,int *xn,int *yn,int *vn,int *nnd,
                    float z);
static void goodneighbor(int x,int y,int v,int xo,int yo,int vo,
                         int *xn,int *yn,int *vn,int *n,int *nnd);
static BOOLEAN vertexdone(int x,int y,int v);
static void setvertexdone(int x,int y,int v);
static BOOLEAN alldone(void);
static void setallnotdone(void);


int one_level(DATAPTR d,int x1,int y1,int x2,int y2,float level)

    {
    int     ix,iy,v;
    int     xt,yt,vt;
    int     nnd;

    sd=d->shotdata;
    in=d->interp;
    dp=d;
    donesize=(unsigned int)d->donesize;
    done=d->done;

    xb1=x1;
    xb2=x2;
    yb1=y1;
    yb2=y2;

    setallnotdone();
    for (iy=y1;iy<=y2;iy++)
        {
        if (bbioskey(1))
            {
            bbioskey(0);
            return(0);
            }
        for (ix=x1;ix<=x2;ix++)
            for (v=1;v<=8;v++)
                {
                if (OUTOFBOUNDS(ix,iy,v) || vertexdone(ix,iy,v))
                    continue;
                if (correct_height(ix,iy,v,level))
                    trace_path(ix,iy,v,level);
                else
                    setvertexdone(ix,iy,v);
                }
        }
    for (iy=y1;iy<y2;iy++)
        {
        if (bbioskey(1))
            {
            bbioskey(0);
            return(0);
            }
        for (ix=x1;ix<x2;ix++)
            {
            if (in[ix][iy]!=level)
                continue;
            if (sd[ix+1][iy]==sd[ix+1][iy+1] && sd[ix+1][iy]==in[ix][iy])
                filltri(ix,iy,1);
            if (sd[ix][iy+1]==sd[ix+1][iy+1] && sd[ix][iy+1]==in[ix][iy])
                filltri(ix,iy,2);
            if (sd[ix][iy]==sd[ix][iy+1] && sd[ix][iy]==in[ix][iy])
                filltri(ix,iy,3);
            if (sd[ix][iy]==sd[ix+1][iy] && sd[ix][iy]==in[ix][iy])
                filltri(ix,iy,4);
            }
        }
    while (!alldone())
        for (iy=y1;iy<=y2;iy++)
            {
            if (bbioskey(1))
                {
                bbioskey(0);
                return(0);
                }
            for (ix=x1;ix<=x2;ix++)
                for (v=1;v<=8;v++)
                    {
                    if (OUTOFBOUNDS(ix,iy,v) || vertexdone(ix,iy,v))
                        continue;
                    VCOPY(xt,yt,vt,ix,iy,v);
                    neighbor(ix,iy,v,&xt,&yt,&vt,&nnd,level);
                    if (nnd<=1)
                        setvertexdone(ix,iy,v);
                    if (nnd)
                        drawline(xpoint(ix,iy,v,level),
                                 ypoint(ix,iy,v,level),
                                 xpoint(xt,yt,vt,level),
                                 ypoint(xt,yt,vt,level));
                    }
            }
    return(1);
    }


static void trace_path(int x,int y,int v,float z)

    {
    int     x2,y2,v2;
    int     xs,ys,vs;
    int     kfp,pc,n,nnd;
    int     closed;

    pc=trace_back(&x,&y,&v,&x2,&y2,&v2,&kfp,z);
    if (pc<0)
        return;
    closed=pc;
    if (!pc && need_count())
        pc=count_points(x,y,v,x2,y2,v2,z);
    VCOPY(xs,ys,vs,x,y,v);
    curve_start(xpoint(x,y,v,z),ypoint(x,y,v,z),closed,pc);
    if (kfp)
        setvertexdone(x,y,v);
    while (1)
        {
        curve_next(xpoint(x2,y2,v2,z),ypoint(x2,y2,v2,z));
        VSWAP(x,y,v,x2,y2,v2);
        n=neighbor(x,y,v,&x2,&y2,&v2,&nnd,z);
        if (n>2)
            {
            if (!nnd)
                setvertexdone(x,y,v);
            break;
            }
        setvertexdone(x,y,v);
        if (!nnd || n<=1)
            break;
        if (closed && xs==x2 && ys==y2 && vs==v2)
            {
            setvertexdone(xs,ys,vs);
            break;
            }
        }
    curve_end();
    }


static int count_points(int x,int y,int v,int x2,int y2,int v2,float z)

    {
    int     n,pc,nnd;

    pc=2;
    while (1)
        {
        VSWAP(x,y,v,x2,y2,v2);
        n=neighbor(x,y,v,&x2,&y2,&v2,&nnd,z);
        if (!n)
            {
            printf("cp: internal error, neighbors=0\n");
            break;
            }
        if (n==1)
            break;
        if (n>2)
            break;
        if (nnd!=2)
            printf("cp: internal error, nnd=%d, should be 2.\n",nnd);
        pc++;
        }
    return(pc);
    }


static int trace_back(int *x,int *y,int *v,int *x2,int *y2,int *v2,int *kfp,
                      float z)

    {
    int     xn,yn,vn;
    int     xs,ys,vs;
    int     n,nnd;
    int     pc;

    (*kfp)=1;
    VCOPY(xn,yn,vn,*x,*y,*v);
    VCOPY(xs,ys,vs,xn,yn,vn);
    (*x2)=-1;
    n=neighbor(xn,yn,vn,x,y,v,&nnd,z);
    if (!n)
        {
        if (single_point(xn,yn,vn))
            putpoint(xn,yn,vn);
        setvertexdone(xn,yn,vn);
        return(-2);
        }
    if (n==1)
        {
        if (!nnd)
            {
            printf("Internal error:  neighbor=1, neighbors not done=0\n");
            setvertexdone(xn,yn,vn);
            return(-2);
            }
        VCOPY(*x2,*y2,*v2,*x,*y,*v);
        VCOPY(*x,*y,*v,xn,yn,vn);
        return(0);
        }
    if (n>2)
        return(-1);
    if (nnd!=2)
        printf("Internal error:  neighbors=2, neighbors not done=%d\n",nnd);
    pc=2;
    while (1)
        {
        VSWAP(xn,yn,vn,*x,*y,*v)
        VCOPY(*x2,*y2,*v2,*x,*y,*v);
        n=neighbor(xn,yn,vn,x,y,v,&nnd,z);
        if (n>2)
            {
            (*kfp)=(nnd<2);
            pc=0;
            break;
            }
        if (n==1)
            {
            if (nnd!=1)
                printf("Internal error2: n=1, nnd=%d\n",nnd);
            pc=0;
            break;
            }
        if (!n)
            {
            pc=0;
            printf("Internal error:  No neighbors in path trace.\n");
            break;
            }
        if ((*x)==xs && (*y)==ys && (*v)==vs)
            {
            (*kfp)=0;
            VCOPY(*x2,*y2,*v2,xn,yn,vn);
            VCOPY(xn,yn,vn,*x,*y,*v);
            break;
            }
        if (nnd!=2)
            printf("Internal error:  nnd=%d, should be 2.\n",nnd);
        pc++;
        }
    VCOPY(*x,*y,*v,xn,yn,vn);
    return(pc);
    }


static BOOLEAN correct_height(int x,int y,int v,float z)

    {
    switch(v)
        {
        case 1:
            return(sd[x][y]==z);
        case 2:
            return((sd[x][y]<z && sd[x][y+1]>z)||(sd[x][y]>z && sd[x][y+1]<z));
        case 3:
            return((sd[x][y]<z && sd[x+1][y]>z)||(sd[x][y]>z && sd[x+1][y]<z));
        case 4:
            return(in[x][y]==z);
        case 5:
            return((sd[x][y]<z && in[x][y]>z)||(sd[x][y]>z && in[x][y]<z));
        case 6:
            return((sd[x+1][y]<z && in[x][y]>z)||(sd[x+1][y]>z && in[x][y]<z));
        case 7:
            return((sd[x][y+1]<z && in[x][y]>z)||(sd[x][y+1]>z && in[x][y]<z));
        case 8:
            return((sd[x+1][y+1]<z&&in[x][y]>z)||(sd[x+1][y+1]>z&&in[x][y]<z));
        }
    return(0);
    }


static float xpoint(int x,int y,int v,float z)

    {
    switch(v)
        {
        case 1:
            return((float)x);
        case 2:
            return((float)x);
        case 3:
            return((float)x+(z-sd[x][y])/(sd[x+1][y]-sd[x][y]));
        case 4:
            return((float)x+.5);
        case 5:
            return((float)x+.5*(z-sd[x][y])/(in[x][y]-sd[x][y]));
        case 6:
            return((float)x+.5+.5*(z-in[x][y])/(sd[x+1][y]-in[x][y]));
        case 7:
            return((float)x+.5*(z-sd[x][y+1])/(in[x][y]-sd[x][y+1]));
        case 8:
            return((float)x+.5+.5*(z-in[x][y])/(sd[x+1][y+1]-in[x][y]));
        }
    return(-1.);
    }


static float ypoint(int x,int y,int v,float z)

    {
    switch(v)
        {
        case 1:
            return((float)y);
        case 2:
            return((float)y+(z-sd[x][y])/(sd[x][y+1]-sd[x][y]));
        case 3:
            return((float)y);
        case 4:
            return((float)y+.5);
        case 5:
            return((float)y+.5*(z-sd[x][y])/(in[x][y]-sd[x][y]));
        case 6:
            return((float)y+.5*(z-sd[x+1][y])/(in[x][y]-sd[x+1][y]));
        case 7:
            return((float)y+.5+.5*(z-in[x][y])/(sd[x][y+1]-in[x][y]));
        case 8:
            return((float)y+.5+.5*(z-in[x][y])/(sd[x+1][y+1]-in[x][y]));
        }
    return(-1.);
    }


static BOOLEAN single_point(int x,int y,int v)

    {
    int     below,above;
    float   z;

    if (v!=1 && v!=4)
        return(0);
    below=above=0;
    if (v==1)
        {
        z=sd[x][y];
        if (y>yb1)
            COUNTPOINT(sd[x][y-1]);
        if (y>yb1 && x>xb1)
            COUNTPOINT(in[x-1][y-1]);
        if (x>xb1)
            COUNTPOINT(sd[x-1][y]);
        if (x>xb1 && y<yb2)
            COUNTPOINT(in[x-1][y]);
        if (y<yb2)
            COUNTPOINT(sd[x][y+1]);
        if (y<yb2 && x<xb2)
            COUNTPOINT(in[x][y]);
        if (x<xb2)
            COUNTPOINT(sd[x+1][y]);
        if (x<xb2 && y>yb1)
            COUNTPOINT(in[x][y-1]);
        }
    else
        {
        z=in[x][y];
        COUNTPOINT(sd[x][y]);
        COUNTPOINT(sd[x+1][y]);
        COUNTPOINT(sd[x][y+1]);
        COUNTPOINT(sd[x+1][y+1]);
        }
    return(below==0 || above==0);
    }


/*
** For neighbor to work properly,
** x,y,v must be a vertex WITHIN the bounds.
*/
static int neighbor(int x,int y,int v,int *xn,int *yn,int *vn,int *nnd,float z)

    {
    int     n;
    int     xo,yo,vo;

    xo=(*xn);
    yo=(*yn);
    vo=(*vn);
    n=0;
    (*nnd)=0;
    switch(v)
        {
        case 1:
            /* Check other vertices */
            if (x>xb1 && y>yb1 && in[x-1][y-1]==z && sd[x][y-1]!=z
                && sd[x-1][y]!=z)
                CKNEIGHBOR(x-1,y-1,4);
            if (x>xb1 && sd[x-1][y]==z && (y==yb2 || in[x-1][y]!=z)
                && (y==yb1 || in[x-1][y-1]!=z))
                CKNEIGHBOR(x-1,y,1);
            if (x>xb1 && y<yb2 && in[x-1][y]==z && sd[x-1][y]!=z
                && sd[x][y+1]!=z)
                CKNEIGHBOR(x-1,y,4);
            if (y<yb2 && sd[x][y+1]==z && (x==xb1 || in[x-1][y]!=z)
                && (x==xb2 || in[x][y]!=z))
                CKNEIGHBOR(x,y+1,1);
            if (y<yb2 && x<xb2 && in[x][y]==z && sd[x][y+1]!=z
                && sd[x+1][y]!=z)
                CKNEIGHBOR(x,y,4);
            if (x<xb2 && sd[x+1][y]==z && (y==yb2 || in[x][y]!=z)
                && (y==yb1 || in[x][y-1]!=z))
                CKNEIGHBOR(x+1,y,1);
            if (x<xb2 && y>yb1 && in[x][y-1]==z && sd[x+1][y]!=z
                && sd[x][y-1]!=z)
                CKNEIGHBOR(x,y-1,4);
            if (y>yb1 && sd[x][y-1]==z && (x==xb1 || in[x-1][y-1]!=z)
                && (x==xb2 || in[x][y-1]!=z))
                CKNEIGHBOR(x,y-1,1);

            /* Check other sides */
            if (x<xb2 && y<yb2 && correct_height(x,y,6,z))
                CKNEIGHBOR(x,y,6);
            if (x<xb2 && y<yb2 && correct_height(x,y,7,z))
                CKNEIGHBOR(x,y,7);
            if (x>xb1 && y<yb2 && correct_height(x-1,y,8,z))
                CKNEIGHBOR(x-1,y,8);
            if (x>xb1 && y<yb2 && correct_height(x-1,y,5,z))
                CKNEIGHBOR(x-1,y,5);
            if (x>xb1 && y>yb1 && correct_height(x-1,y-1,7,z))
                CKNEIGHBOR(x-1,y-1,7);
            if (x>xb1 && y>yb1 && correct_height(x-1,y-1,6,z))
                CKNEIGHBOR(x-1,y-1,6);
            if (x<xb2 && y>yb1 && correct_height(x,y-1,5,z))
                CKNEIGHBOR(x,y-1,5);
            if (x<xb2 && y>yb1 && correct_height(x,y-1,8,z))
                CKNEIGHBOR(x,y-1,8);
            break;
        case 2:
            if (x>xb1 && correct_height(x-1,y,6,z))
                CKNEIGHBOR(x-1,y,6);
            if (x>xb1 && correct_height(x-1,y,8,z))
                CKNEIGHBOR(x-1,y,8);
            if (x<xb2 && correct_height(x,y,5,z))
                CKNEIGHBOR(x,y,5);
            if (x<xb2 && correct_height(x,y,7,z))
                CKNEIGHBOR(x,y,7);
            if (x>xb1 && in[x-1][y]==z)
                CKNEIGHBOR(x-1,y,4);
            if (x<xb2 && in[x][y]==z)
                CKNEIGHBOR(x,y,4);
            break;
        case 3:
            if (y<yb2 && correct_height(x,y,5,z))
                CKNEIGHBOR(x,y,5);
            if (y<yb2 && correct_height(x,y,6,z))
                CKNEIGHBOR(x,y,6);
            if (y>yb1 && correct_height(x,y-1,7,z))
                CKNEIGHBOR(x,y-1,7);
            if (y>yb1 && correct_height(x,y-1,8,z))
                CKNEIGHBOR(x,y-1,8);
            if (y>yb1 && in[x][y-1]==z)
                CKNEIGHBOR(x,y-1,4);
            if (y<yb2 && in[x][y]==z)
                CKNEIGHBOR(x,y,4);
            break;
        case 4:
            /* Check other vertices */
            if (sd[x][y]==z && sd[x][y+1]!=z && sd[x+1][y]!=z)
                CKNEIGHBOR(x,y,1);
            if (sd[x+1][y]==z && sd[x][y]!=z && sd[x+1][y+1]!=z)
                CKNEIGHBOR(x+1,y,1);
            if (sd[x][y+1]==z && sd[x][y]!=z && sd[x+1][y+1]!=z)
                CKNEIGHBOR(x,y+1,1);
            if (sd[x+1][y+1]==z && sd[x][y+1]!=z && sd[x+1][y]!=z)
                CKNEIGHBOR(x+1,y+1,1);

            /* Check other sides */
            if (correct_height(x+1,y,2,z))
                CKNEIGHBOR(x+1,y,2);
            if (correct_height(x,y+1,3,z))
                CKNEIGHBOR(x,y+1,3);
            if (correct_height(x,y,2,z))
                CKNEIGHBOR(x,y,2);
            if (correct_height(x,y,3,z))
                CKNEIGHBOR(x,y,3);
            break;
        case 5:
            if (correct_height(x,y,2,z))
                CKNEIGHBOR(x,y,2);
            if (correct_height(x,y,7,z))
                CKNEIGHBOR(x,y,7);
            if (correct_height(x,y,6,z))
                CKNEIGHBOR(x,y,6);
            if (correct_height(x,y,3,z))
                CKNEIGHBOR(x,y,3);
            if (sd[x][y+1]==z)
                CKNEIGHBOR(x,y+1,1);
            if (sd[x+1][y]==z)
                CKNEIGHBOR(x+1,y,1);
            break;
        case 6:
            if (correct_height(x+1,y,2,z))
                CKNEIGHBOR(x+1,y,2);
            if (correct_height(x,y,5,z))
                CKNEIGHBOR(x,y,5);
            if (correct_height(x,y,8,z))
                CKNEIGHBOR(x,y,8);
            if (correct_height(x,y,3,z))
                CKNEIGHBOR(x,y,3);
            if (sd[x][y]==z)
                CKNEIGHBOR(x,y,1);
            if (sd[x+1][y+1]==z)
                CKNEIGHBOR(x+1,y+1,1);
            break;
        case 7:
            if (correct_height(x,y,2,z))
                CKNEIGHBOR(x,y,2);
            if (correct_height(x,y,5,z))
                CKNEIGHBOR(x,y,5);
            if (correct_height(x,y,8,z))
                CKNEIGHBOR(x,y,8);
            if (correct_height(x,y+1,3,z))
                CKNEIGHBOR(x,y+1,3);
            if (sd[x][y]==z)
                CKNEIGHBOR(x,y,1);
            if (sd[x+1][y+1]==z)
                CKNEIGHBOR(x+1,y+1,1);
            break;
        case 8:
            if (correct_height(x+1,y,2,z))
                CKNEIGHBOR(x+1,y,2);
            if (correct_height(x,y,7,z))
                CKNEIGHBOR(x,y,7);
            if (correct_height(x,y,6,z))
                CKNEIGHBOR(x,y,6);
            if (correct_height(x,y+1,3,z))
                CKNEIGHBOR(x,y+1,3);
            if (sd[x+1][y]==z)
                CKNEIGHBOR(x+1,y,1);
            if (sd[x][y+1]==z)
                CKNEIGHBOR(x,y+1,1);
            break;
        }
    return(n);
    }


static void goodneighbor(int x,int y,int v,int xo,int yo,int vo,
                         int *xn,int *yn,int *vn,int *n,int *nnd)

    {
    (*n)=(*n)+1;
    if (vertexdone(x,y,v))
        return;
    (*nnd)=(*nnd)+1;
    if (x!=xo || y!=yo || v!=vo)
        {
        (*xn)=x;
        (*yn)=y;
        (*vn)=v;
        }
    }


static BOOLEAN vertexdone(int x,int y,int v)

    {
    return(dp->done[x*dp->ysize+y]&(1<<(v-1)));
    }


static void setvertexdone(int x,int y,int v)

    {
    if (!vertexdone(x,y,v))
        {
        dp->done[x*dp->ysize+y] |= (1<<(v-1));
        vdone++;
        }
    }


static BOOLEAN alldone(void)

    {
    return(vdone==vmax);
    }


static void setallnotdone(void)

    {
    unsigned int     i;

    for (i=0;i<dp->donesize;i++)
        dp->done[i]=0;
    vmax=(long)(xb2-xb1)*(yb2-yb1)*8L+2L*(xb2-xb1)+2L*(yb2-yb1)+1;
    /*
    vmax=(long)(dp->xsize-1)*(dp->ysize-1)*8L+2L*(dp->xsize-1)
         +2L*(dp->ysize-1)+1;
    */
    vdone=0;
    }
