/*
**
** CLPLOT.C    Plots contour lines.
**
** Will Menninger, 11/92, programmed in Turbo C 2.0
** Original logic done by Danny Katz.
** Logic totally redone 1-16-93 so that splines could be used.
**
*/

#include "contour.h"
#include <ps.h>

#define ISWAP(x,y)      {int t; t=(x);(x)=(y);(y)=t;}
#define AMAX3(x,y,z)    (x>y?(x>z ? x : z):(y>z ? y : z))

/* VGA parameters */
#define LMAR        140
#define RMAR        20
#define TOPMAR      28
#define BOTMAR      25
#define LEGMAR      60
#define LCFACTOR    30
#define x_vga(x)    (int)(xoff_vga+(x)*xscale_vga)
#define y_vga(y)    (int)(yoff_vga+(y)*yscale_vga)

/* PostScript parameters */
#define LTHICK      .96
#define CTHICK      .72
#define AXTICK      0.012
#define AXTMAR      0.008
#define YAXNMAR     4.
#define XAXNMAR     .5
#define TMAR        1.
#define LWIDTH      30.
#define LSPACE      8.
#define LNMAR       4.
#define MAXLABDEN   4
#define TOOCLOSE    0.4
#define LABMAX      20
#define POINTBUF    7
#define PSSCALE     10.
#define x_ps(x)     (xoff_ps+(x)*xscale_ps)
#define y_ps(y)     (yoff_ps+(y)*yscale_ps)

typedef struct labellist {
               float x1[LABMAX];
               float x2[LABMAX];
               float y1[LABMAX];
               float y2[LABMAX];
               float level[LABMAX];
               int   color[LABMAX];
               struct labellist *next;
               } LABLST;

/* Globals common to VGA and PostScript */
static int      ccolor;
static int      legend;
static float    legendy;
static float    level;
static DATAPTR  dp;
static int      postscript;
static int      xb1,xb2,yb1,yb2;
static float    axtmar,axtick;

/* VGA globals */
static int      vgax0,vgay0;
static float    xoff_vga,yoff_vga;
static float    xscale_vga,yscale_vga;

/* PostScript globals */
static float    xoff_ps,yoff_ps;
static float    xscale_ps,yscale_ps;
static float    psx1,psy1,psx2,psy2;
static float    pstop,psbottom,psleft,psright;
static float    fsize;
static char     psfilename[80];
static int      xpoint[POINTBUF];
static int      ypoint[POINTBUF];
static int      psindex;
static int      pscs;
static int      append,eject;

/* Globels dealing with labeling */
static int      ldx,ldy;
static float    labelx[MAXLABDEN][MAXLABDEN];
static float    labely[MAXLABDEN][MAXLABDEN];
static float    gradient[MAXLABDEN][MAXLABDEN];
static int      labelcount;
static float    labeldx;
static LABLST  *labels;

/* Curve drawing globals */
static int      cindex;
static float    cax[5];
static float    cay[5];
static int      closed;
static int      cpoints;

static int      pscolor;


static void vga_setup(int x1,int y1,int x2,int y2);
static void ps_setup(int x1,int y1,int x2,int y2);
static void ps_start(int grids);
static void draw_all_contours(int x1,int y1,int x2,int y2,char *buf);
static void finish_all_contours(int status);
static int contourloop(int x1,int y1,int x2,int y2,float start,
                       float end,float step);
static int do_one_contour(int x1,int y1,int x2,int y2,float lev);
static void setgrey(int ccolor);
static void pssetcolor(int cnum);
static void vgafilltri(float x1,float y1,float x2,float y2,float x3,float y3);
static void clearout(void);
static void checklabel(float x,float y);
static float calcgrad(float x,float y);
static void addlabel(float x,float y,float dx,float level);
static void dellabel(float x,float y,float level);
static void printlabels(void);
static int toocrowded(float x,float y,float dx);
static void freelabels(void);
static void freelastlabel(void);


void contour_plot(DATAPTR d)

    {
    int     i,grids_on;
    char    buf[80];
    char    prompt[100];
    int     x1,y1,x2,y2;
    float   xmin,xmax,ymin,ymax;
    int     not_clear,in,color;

    if (!loaded(TRUE))
        return;
    dp=d;
    if (!get_range('X',d,&xmin,&xmax) || !get_range('Y',d,&ymin,&ymax))
        return;
    x1=xindex(d,xmin)+.0001;
    x2=xindex(d,xmax)+.99;
    y1=yindex(d,ymin)+.0001;
    y2=yindex(d,ymax)+.99;
    if (x2<0 || y2<0 || x1>d->xsize-1 || y1>d->ysize-1)
        return;
    if (x1<0)
        x1=0;
    if (y1<0)
        y1=0;
    if (x2>d->xsize-1)
        x2=d->xsize-1;
    if (y2>d->ysize-1)
        y2=d->ysize-1;
    if (x1==x2 || y1==y2)
        return;
    normal("Enter label density (0 for legend, Def=2,2):  ");
    if (bgetinput("",buf,10)==ESCAPE)
        return;
    normal("\r\n\r\n");
    if (buf[0]==EOS)
        i=0;
    else
        i=sscanf(buf,"%d,%d",&ldx,&ldy);
    if (!i)
        ldx=ldy=2;
    else if (i==1)
        ldy=ldx;
    if (ldy<0)
        ldy=0;
    if (ldy>MAXLABDEN)
        ldy=MAXLABDEN;
    if (ldx<0)
        ldx=0;
    if (ldx>MAXLABDEN)
        ldx=MAXLABDEN;
    legend=(ldx==0 || ldy==0);

    normal("Enter contour values.  Prefix a \"/\" for step values.\r\n");
    normal("E.g. \"1.5 3 6 /2 12\" will do values 1.5, 3, 6, 8, 10, and 12.\r\n");
    if (bgetinput("Values:  ",buf,65)==ESCAPE)
        return;

    xb1=x1;
    yb1=y1;
    xb2=x2;
    yb2=y2;
    vga_setup(x1,y1,x2,y2);
    ps_setup(x1,y1,x2,y2);

    /* Draw contours */
    postscript=0;
    draw_all_contours(x1,y1,x2,y2,buf);

    /* Toggle between color and grey-scale if VGA */
    grids_on=0;
    putgrid(0);
    pscolor=color=1;
    not_clear=0;
    postscript=1;
    strcpy(prompt,"<T> color/bw, <G> grid, <P> print, <S> save, <ESC> exit");
    ps_update(0,10,prompt,NULL);
    while (1)
        {
        in=bbioskey(0)&0xff;
        if (not_clear)
            {
            not_clear=0;
            ps_clr_update();
            }
        if (in=='s' || in=='S')
            {
            ps_update(0,15,vgasave(),NULL);
            not_clear=1;
            continue;
            }
        if (in=='p' || in=='P')
            {
            if (nextname(10,psfilename,&append,&eject))
                {
                postscript=1;
                ps_start(!i);
                draw_all_contours(x1,y1,x2,y2,buf);
                postscript=0;
                not_clear=1;
                }
            else
                ps_update(0,10,prompt,NULL);
            continue;
            }
        if (canswitch() && (in=='t' || in=='T'))
            {
            color=!color;
            pscolor=color;
            contour_colors(color);
            continue;
            }
        if (in=='g' || in=='G')
            {
            if (grids_on && erase_grids())
                grids_on=0;
            else
                {
                putgrid(1);
                grids_on=1;
                }
            continue;
            }
        break;
        }
    free_grids();
    graphics_off();
    }


static void vga_setup(int x1,int y1,int x2,int y2)

    {
    int     x0,y0,w,h;

    if (graphics_on()!=NO_ERROR)
        return;
    contour_colors(1);
    if (legend)
        {
        w=getmaxx()-LMAR-RMAR;
        x0=LMAR;
        }
    else
        {
        w=getmaxx()-LMAR-RMAR+LEGMAR;
        x0=LMAR-LEGMAR;
        }
    h=getmaxy()-TOPMAR-BOTMAR;
    y0=getmaxy()-BOTMAR;
    vgax0=x0;
    vgay0=y0;
    xscale_vga=(float)w/(float)(x2-x1);
    yscale_vga=-(float)h/(float)(y2-y1);
    xoff_vga=(float)x0-(float)x1*xscale_vga;
    yoff_vga=(float)y0-(float)y1*yscale_vga;

    setgriddim(dp->xmin+x1*(dp->xmax-dp->xmin)/(dp->xsize-1.),
               dp->ymin+y1*(dp->ymax-dp->ymin)/(dp->ysize-1.),
               dp->xmin+x2*(dp->xmax-dp->xmin)/(dp->xsize-1.),
               dp->ymin+y2*(dp->ymax-dp->ymin)/(dp->ysize-1.),
               x0-1,y0+1,x0+w,y0-h,"%g","%g");

    put_titles(dp,320,10,x0+w/2.,y0+1,x0,y0-h/2.);
    if (legend)
        {
        settextjustify(LEFT_TEXT,BOTTOM_TEXT);
        outtextxy(0,TOPMAR-5,dp->ztitle);
        }


    /* Bounding rectangle */
    setfillstyle(SOLID_FILL,1);
    bar(x0-1,y0+1,x0+w,y0-h);
/*
{
int     i,j;
setcolor(2);
for (i=xb1;i<xb2;i++)
for (j=yb1;j<yb2;j++)
{
moveto(x_vga(i),y_vga(j));
lineto(x_vga(i+1.),y_vga(j));
lineto(x_vga(i+1.),y_vga(j+1.));
lineto(x_vga(i),y_vga(j+1.));
lineto(x_vga(i),y_vga(j));
lineto(x_vga(i+1.),y_vga(j+1.));
moveto(x_vga(i),y_vga(j+1.));
lineto(x_vga(i+1.),y_vga(j));
}
}
*/
    bbox();
    }


static void ps_setup(int x1,int y1,int x2,int y2)

    {
    float   totalheight,totalwidth;
    float   fx1,fy1,fx2,fy2;
    float   pagex,pagey;

    if (dp->pagex<0)
        {
        pagex=5.0;
        pagey=5.0;
        }
    else
        {
        pagex=dp->pagex;
        pagey=dp->pagey;
        }
    fsize=fontsize(pagex,pagey);
    axtick=AXTICK*(pagex+pagey)*36.;
    axtmar=AXTMAR*(pagex+pagey)*36.;
    totalheight=pagey*72.+(3.+TMAR+XAXNMAR)*fsize+axtick+axtmar;
    if (!legend)
        totalwidth=pagex*72.+(YAXNMAR+1.)*fsize+axtick+axtmar;
    else
        totalwidth=pagex*72.+(YAXNMAR+LNMAR+1.)*fsize+LWIDTH+LSPACE+
                   axtick+axtmar;
    psbottom=dp->pagebotmar<-90. ? 11.*36.-totalheight/2. : dp->pagebotmar*72.;
    psleft=dp->pageleftmar<-90. ? 8.5*36.-totalwidth/2. : dp->pageleftmar*72.;
    psright=psleft+totalwidth;
    pstop=psbottom+totalheight;
    psx1=psright-pagex*72.;
    psy1=psbottom+(2.+XAXNMAR)*fsize+axtick+axtmar;
    psx2=psright;
    psy2=psy1+pagey*72.;
    fx1=dp->xmin+x1*(dp->xmax-dp->xmin)/(dp->xsize-1.);
    fy1=dp->ymin+y1*(dp->ymax-dp->ymin)/(dp->ysize-1.);
    fx2=dp->xmin+x2*(dp->xmax-dp->xmin)/(dp->xsize-1.);
    fy2=dp->ymin+y2*(dp->ymax-dp->ymin)/(dp->ysize-1.);
    ps_setgriddim(fx1,fy1,fx2,fy2,psx1,psy1,psx2,psy2,
                  fsize,axtick,axtmar,"%g","%g");
    xscale_ps=(psx2-psx1)/(float)(x2-x1);
    yscale_ps=(psy2-psy1)/(float)(y2-y1);
    xoff_ps=psx1-(float)x1*xscale_ps;
    yoff_ps=psy1-(float)y1*yscale_ps;
    }


static void ps_start(int grids)

    {
    int     i;
    char    buf[100];
    float   dx,dy;

    if (append)
        g_append(psfilename);
    else
        g_init(psfilename);
    dx=xscale_ps*PSSCALE;
    dy=yscale_ps*PSSCALE;
    g_psout("/F { /ntri exch def\n");
    sprintf(buf,"   %d sub %g mul %g add /yp exch def\n",yb1,dy,psy1*PSSCALE);
    g_psout(buf);
    sprintf(buf,"   %d sub %g mul %g add /xp exch def\n",xb1,dx,psx1*PSSCALE);
    g_psout(buf);
    sprintf(buf,"   ntri 1 eq {xp %g add yp moveto\n",dx);
    g_psout(buf);
    sprintf(buf,"   xp %g add yp %g add lineto\n",dx,dy);
    g_psout(buf);
    sprintf(buf,"   xp %g add yp %g add lineto}\n",dx/2.,dy/2.);
    g_psout(buf);
    sprintf(buf,"   {ntri 2 eq {xp %g add yp %g add moveto\n",dx,dy);
    g_psout(buf);
    sprintf(buf,"   xp yp %g add lineto\n",dy);
    g_psout(buf);
    sprintf(buf,"   xp %g add yp %g add lineto}\n",dx/2.,dy/2.);
    g_psout(buf);
    sprintf(buf,"   {ntri 3 eq {xp yp moveto xp yp %g add lineto\n",dy);
    g_psout(buf);
    sprintf(buf,"   xp %g add yp %g add lineto}\n",dx/2.,dy/2.);
    g_psout(buf);
    sprintf(buf,"   {xp yp moveto xp %g add yp lineto\n",dx);
    g_psout(buf);
    sprintf(buf,"   xp %g add yp %g add lineto}\n",dx/2.,dy/2.);
    g_psout(buf);
    g_psout("   ifelse } ifelse } ifelse\n");
    g_psout("   closepath fill } def\n");

    g_psout("/P { /nvtx exch def\n");
    sprintf(buf,"   %d sub %g mul %g add /yp exch def\n",yb1,dy,psy1*PSSCALE);
    g_psout(buf);
    sprintf(buf,"   %d sub %g mul %g add /xp exch def\n",xb1,dx,psx1*PSSCALE);
    g_psout(buf);
    sprintf(buf,"   nvtx 4 eq {xp %g add yp %g add moveto\n",dx/2.,dy/2.);
    g_psout(buf);
    g_psout("   currentpoint lineto}\n");
    g_psout("   {xp yp moveto xp yp lineto} ifelse stroke } def\n");

    g_psout("/M {");
    for (i=0;i<POINTBUF-1;i++)
        g_psout("lineto ");
    g_psout("\n    lineto ST} def\n");
    g_psout("/N {/np exch def 1 1 np {pop lineto} for ST} def\n");
    g_psout("/S {/np exch def moveto 2 1 np {pop lineto} for ST} def\n");
    sprintf(buf,"\\fTR\\s{%g}",fsize);
    g_showtext(buf);
    g_linecap(2);
    ps_putgrid(grids);
    if (dp->no_title_margin)
        ps_put_titles(dp,(psx1+psx2)/2.,psy2+fsize/3.,
                         (psx1+psx2)/2.,psbottom+fsize*1.4,
                         psx1-axtmar-YAXNMAR*fsize*3./5.,(psy1+psy2)/2.);
    else
        ps_put_titles(dp,(psleft+psright)/2.,pstop+fsize,
                         (psx1+psx2)/2.,psbottom+fsize,
                         psx1-axtmar-YAXNMAR*fsize,(psy1+psy2)/2.);
    if (legend)
        {
        g_justify(6);
        g_moveto(psleft,psy2+fsize/2.);
        g_showtext(dp->ztitle);
        }
    }


static void draw_all_contours(int x1,int y1,int x2,int y2,char *buf)

    {
    int     i,j,ls,stepval,notstarted;
    float   start,fstep;
    char    buf2[80];
    float   val;
    int     status;

    labelcount=0;
    ccolor=(postscript && !pscolor) ? 0 : 6;
    if (postscript)
        {
        g_linecap(1);
        g_linejoin(1);
        g_linewidth(CTHICK);
        }
    legendy=postscript ? psy2-fsize/2. : TOPMAR;
    if (buf[0]==EOS)
        {
        float   min,max;
        float   offset,pertick;
        int     dummy,first,last;

        min=dataset_min(dp,x1,y1,x2,y2);
        max=dataset_max(dp,x1,y1,x2,y2);
        getticks(min,max,&offset,&pertick,&first,&last,&dummy,10);
        offset=offset+first*pertick;
        status=contourloop(x1,y1,x2,y2,offset,offset+(last-first)*pertick,
                           pertick);
        finish_all_contours(status);
        return;
        }
    i=0;
    ls=0;
    notstarted=1;
    start=-1;
    fstep=1.;
    stepval=0;
    while (1)
        {
        for (;buf[i]==' ' || buf[i]==',' || buf[i]=='\t';i++);
        if (buf[i]=='/')
            {
            i++;
            stepval=1;
            }
        else
            stepval=0;
        if (buf[i]==EOS)
            break;
        for (j=0;buf[i]!=' ' && buf[i]!=',' && buf[i]!='\t' && buf[i]!=EOS;i++)
            buf2[j++]=buf[i];
        if (!j)
            if (buf[i]==EOS)
                break;
            else
                continue;
        buf2[j]=EOS;
        val=atof(buf2);
        if (!stepval && !ls)
            {
            if (!notstarted)
                if (!(status=contourloop(x1,y1,x2,y2,start,start,0.)))
                    {
                    ls=0;
                    notstarted=1;
                    break;
                    }
            start=val;
            notstarted=0;
            continue;
            }
        if (!stepval && ls)
            {
            if (!(status=contourloop(x1,y1,x2,y2,start,val,fstep)))
                {
                ls=0;
                notstarted=1;
                break;
                }
            notstarted=1;
            ls=0;
            continue;
            }
        if (stepval && !ls)
            {
            if (notstarted)
                start=dp->min;
            fstep=val;
            ls=1;
            continue;
            }
        if (stepval && ls)
            {
            if (!(status=contourloop(x1,y1,x2,y2,start,
                                      fstep<0 ? dp->min : dp->max,fstep)))
                {
                ls=0;
                notstarted=1;
                break;
                }
            start=dp->min;
            notstarted=0;
            fstep=val;
            continue;
            }
        }
    if (!ls && !notstarted)
        status=contourloop(x1,y1,x2,y2,start,start,0.);
    else
        if (ls)
            status=contourloop(x1,y1,x2,y2,start,fstep<0. ? dp->min : dp->max,
                               fstep);
    finish_all_contours(status);
    }


static void finish_all_contours(int status)

    {
    if (postscript)
        g_grey(0.);
    if (status && !legend)
        printlabels();
    freelabels();
    if (postscript)
        {
        ps_bbox();
        if (eject)
            g_showpage();
        g_close();
        if (!status)
            {
            ps_update(1,12,"PostScript graph interrupted.",NULL);
            finalprint(psfilename,0);
            }
        else
            finalprint(psfilename,1);
        }
    }


static int contourloop(int x1,int y1,int x2,int y2,float start,
                       float end,float step)

    {
    int     i,n;
    float   z;

    if (step==0)
        return(do_one_contour(x1,y1,x2,y2,start));
    else
        {
        n=(end-start)/step+.001;
        if (!n)
            return(do_one_contour(x1,y1,x2,y2,start));
        else
            for (i=0;i<=n;i++)
                {
                z=start+i*step;
                if (fabs(z/step)<1e-4)
                    z=0;
                if (!do_one_contour(x1,y1,x2,y2,z))
                    return(0);
                }
        }
    return(1);
    }


static int do_one_contour(int x1,int y1,int x2,int y2,float lev)

    {
    char    buf[80];
    int     x,y;

    level=lev;
    if (!postscript)
        {
        if (legend)
            {
            setcolor(ccolor);
            settextjustify(LEFT_TEXT,CENTER_TEXT);
            line(0,(int)legendy+4,32,(int)legendy+4);
            sprintf(buf,"%g",lev);
            setcolor(15);
            outtextxy(36,(int)legendy+4,buf);
            legendy+=10;
            }
        setcolor(ccolor);
        setfillstyle(SOLID_FILL,ccolor);
        }
    else
        {
        ps_update(0,15,"Doing contour level %g",&lev);
        if (legend)
            {
            setgrey(ccolor);
            g_justify(3);
            g_linewidth(LTHICK);
            g_line(psleft,legendy,psleft+LWIDTH,legendy);
            g_moveto(psleft+LWIDTH+LSPACE,legendy);
            g_grey(0.);
            sprintf(buf,"%g",lev);
            g_showtext(buf);
            legendy-=fsize;
            }
        setgrey(ccolor);
        }

    for (x=0;x<ldx;x++)
        for (y=0;y<ldy;y++)
            gradient[x][y]=-1;
    sprintf(buf,"%g",lev);
    labeldx=fsize*strlen(buf)*.3;

    if (postscript)
        {
        g_scale(1/PSSCALE,1/PSSCALE);
        g_linewidth(CTHICK*PSSCALE);
        }

    if (!one_level(dp,x1,y1,x2,y2,lev))
        {
        if (postscript)
            {
            g_unscale();
            g_linewidth(CTHICK);
            }
        return(0);
        }

    if (postscript)
        {
        g_unscale();
        g_linewidth(CTHICK);
        }

    if (!legend)
        clearout();

    ccolor++;
    if (!postscript && ccolor>9)
        ccolor=2;
    if (pscolor && postscript && ccolor>9)
        ccolor=2;
    return(1);
    }


int need_count(void)

    {
    return(0);
    }


void curve_start(float x,float y,int cc,int count)

    {
    cax[0]=x;
    cay[0]=y;
    cax[4]=x;
    cay[4]=y;
    cindex=1;
    closed=cc;
    if (postscript)
        {
        xpoint[0]=x_ps(x)*PSSCALE+.5;
        ypoint[0]=y_ps(y)*PSSCALE+.5;
        psindex=1;
        pscs=1;
        }
    else
        moveto(x_vga(x),y_vga(y));
    cpoints=count;
    }


void curve_next(float x,float y)

    {
    int     i;
    char    buf[100];

    if (cindex<3)
        {
        cax[cindex]=x;
        cax[cindex]=y;
        }
    cindex++;
    if (postscript)
        {
        xpoint[psindex]=x_ps(x)*PSSCALE+.5;
        ypoint[psindex]=y_ps(y)*PSSCALE+.5;
        psindex++;
        if (psindex>=POINTBUF)
            {
            for (i=POINTBUF-1;i>=0;i--)
                {
                sprintf(buf,"%d %d ",xpoint[i],ypoint[i]);
                g_psout(buf);
                }
            if (pscs)
                {
                sprintf(buf,"%d S\n",POINTBUF);
                g_psout(buf);
                pscs=0;
                }
            else
                g_psout("M\n");
            psindex=0;
            }
        }
    else
        lineto(x_vga(x),y_vga(y));
    checklabel((x+cax[4])/2.,(y+cay[4])/2.);
    cax[4]=x;
    cay[4]=y;
    }


void curve_end(void)

    {
    int     i;
    char    buf[100];

    if (postscript)
        {
        if (closed)
            {
            xpoint[psindex]=x_ps(cax[0])*PSSCALE+.5;
            ypoint[psindex]=y_ps(cay[0])*PSSCALE+.5;
            psindex++;
            }
        if (psindex)
            {
            for (i=psindex-1;i>=0;i--)
                {
                sprintf(buf,"%d %d ",xpoint[i],ypoint[i]);
                g_psout(buf);
                }
            sprintf(buf,"%d %s\n",psindex,pscs ? "S" : "N");
            g_psout(buf);
            pscs=0;
            psindex=0;
            }
        }
    if (closed)
        {
        if (!postscript)
            lineto(x_vga(cax[0]),y_vga(cay[0]));
        checklabel((cax[0]+cax[4])/2.,(cay[0]+cay[4])/2.);
        }
    }


static void setgrey(int ccolor)

    {
    if (pscolor)
        {
        pssetcolor(ccolor);
        return;
        }
    switch(ccolor%4)
        {
        case 0:
            g_grey(0.);
            break;
        case 1:
            g_grey(.3);
            break;
        case 2:
            g_grey(.6);
            break;
        case 3:
            g_grey(.8);
            break;
        }
    }


void contour_colors(int color)

    {
    static int cmap[11][4]={{80,80,80,10},{0,0,0,1},{0,255,0,1},{0,255,255,1},
                           {0,50,255,1},{150,0,255,1},{255,0,255,1},
                           {255,130,0,1},{255,255,0,1},{255,255,255,1},
                           {255,255,255,3}};
    int        i,j,n,r,g,b;

    if (!color)
        {
        rgbset(0,0,255,0);
        n= canswitch() ? 15 : 4;      /* canswitch() == 1 if VGA */
        for (i=0;i<n;i++)
            {
            j=i*255/(n-1);
            rgbset(j,j,j,i+1);
            }
        }
    else
        {
        rgbset(cmap[0][0],cmap[0][1],cmap[0][2],0);
        n=1;
        for (i=1;i<=cmap[0][3];i++)
            for (j=1;j<=cmap[i][3];j++)
                {
                r=cmap[i-1][0]+(cmap[i][0]-cmap[i-1][0])*j/cmap[i][3];
                g=cmap[i-1][1]+(cmap[i][1]-cmap[i-1][1])*j/cmap[i][3];
                b=cmap[i-1][2]+(cmap[i][2]-cmap[i-1][2])*j/cmap[i][3];
                rgbset(r,g,b,n++);
                }
        }
    }


static void pssetcolor(int cnum)

    {
    static float cmap[10][4]={{.5,.5,.5,9},{0,0,0,1},{0,1,0,1},{0,1,1,1},
                           {0,.6,1,1},{.6,0,1,1},{1,0,1,1},
                           {1,.6,0,1},
                           {.7,.4,0,1},{0,0,0,1}};
    int        i,j,n;
    float      r,g,b;

    if (cnum==0)
        {
        g_rgb(0.,0.,0.);
        return;
        }
    n=1;
    for (i=1;i<=cmap[0][3];i++)
        for (j=1;j<=cmap[i][3];j++)
            {
            r=cmap[i-1][0]+(cmap[i][0]-cmap[i-1][0])*j/cmap[i][3];
            g=cmap[i-1][1]+(cmap[i][1]-cmap[i-1][1])*j/cmap[i][3];
            b=cmap[i-1][2]+(cmap[i][2]-cmap[i-1][2])*j/cmap[i][3];
            if (n==cnum)
                {
                g_rgb(r,g,b);
                return;
                }
            n++;
            }
    }


void drawline(float x1,float y1,float x2,float y2)

    {
    char    buf[100];

    if (postscript)
        {
        sprintf(buf,"%d %d MT %d %d LT ",(int)(x_ps(x1)*PSSCALE+.5),
                (int)(y_ps(y1)*PSSCALE+.5),
                (int)(x_ps(x2)*PSSCALE+.5),
                (int)(y_ps(y2)*PSSCALE+.5));
        g_psout(buf);
        g_stroke();
        }
    else
        {
/*
int i;
i=getcolor();
setcolor(15);
*/
        line(x_vga(x1),y_vga(y1),x_vga(x2),y_vga(y2));
/*
setcolor(i);
*/
        }
    }


void filltri(int x,int y,int nt)

    {
    char    buf[40];

    if (postscript)
        {
        sprintf(buf,"%d %d %d F\n",x,y,nt);
        g_psout(buf);
        }
    switch (nt)
        {
        case 1:
            vgafilltri(x+1.,(float)y,x+1.,y+1.,x+.5,y+.5);
            break;
        case 2:
            vgafilltri(x+1.,y+1.,(float)x,y+1.,x+.5,y+.5);
            break;
        case 3:
            vgafilltri((float)x,(float)y,(float)x,y+1.,x+.5,y+.5);
            break;
        case 4:
            vgafilltri((float)x,(float)y,x+1.,(float)y,x+.5,y+.5);
            break;
        }
    }


static void vgafilltri(float x1,float y1,float x2,float y2,float x3,float y3)

    {
    int points[6];

    if (!postscript)
        {
        points[0] = x_vga(x1);
        points[1] = y_vga(y1);
        points[2] = x_vga(x2);
        points[3] = y_vga(y2);
        points[4] = x_vga(x3);
        points[5] = y_vga(y3);
        fillpoly(3,points);
        }
    checklabel((x1+x2+x3)/3.,(y1+y2+y3)/3.);
    }


void putpoint(int x,int y,int v)

    {
    char    buf[20];
    float   x1,y1;

    if (postscript)
        {
        sprintf(buf,"%d %d %d P\n",x,y,v);
        g_psout(buf);
        }
    switch (v)
        {
        case 1:
            x1=x;
            y1=y;
            break;
        case 2:
            x1=x;
            y1=y+.5;
            break;
        case 3:
            x1=x+.5;
            y1=y;
            break;
        case 4:
            x1=x+.5;
            y1=y+.5;
            break;
        case 5:
            x1=x+.25;
            y1=y+.25;
            break;
        case 6:
            x1=x+.75;
            y1=y+.25;
            break;
        case 7:
            x1=x+.25;
            y1=y+.75;
            break;
        case 8:
            x1=x+.75;
            y1=y+.75;
            break;
        default:
            x1=x;
            y1=y;
            break;
        }
    checklabel(x1,y1);
    if (!postscript)
        {
        putpixel(x_vga(x1),y_vga(y1),ccolor);
/*
struct fillsettingstype x;
getfillsettings((struct fillsettingstype far *)&x);
        setfillstyle(SOLID_FILL,4);
        bar(x_vga(x1)-2,y_vga(y1)-1,x_vga(x1)+2,y_vga(y1)+1);
setfillstyle(x.pattern,x.color);
*/
        }
    }


static void clearout(void)

    {
    int     count[MAXLABDEN][MAXLABDEN];
    int     i,j,x,y;
    float   dx,dy;
    float   dtc;
    float   xdist,ydist;

    xdist=(psx2-psx1)/ldx;
    ydist=(psy2-psy1)/ldy;
    dtc=TOOCLOSE/2.0*(xdist*xdist+ydist*ydist);
    while (1)
        {
        for (x=0;x<ldx;x++)
            for (y=0;y<ldy;y++)
                count[x][y]=0;
        for (x=0;x<ldx;x++)
            for (y=0;y<ldy;y++)
                {
                if (gradient[x][y]<0)
                    continue;
                for (i=0;i<ldx;i++)
                    for (j=0;j<ldy;j++)
                        {
                        if (i==x && j==y)
                            continue;
                        if (gradient[i][j]<0)
                            continue;
                        dx=(labelx[x][y]-labelx[i][j])*xscale_ps;
                        dy=(labely[x][y]-labely[i][j])*yscale_ps;
                        if (dx*dx+dy*dy<dtc)
                            count[x][y]++;
                        }
                }
        i=-1;
        j=-1;
        for (x=0;x<ldx;x++)
            for (y=0;y<ldy;y++)
                if (count[x][y] && (i==-1 || count[x][y]>count[i][j]
                    || (count[x][y]==count[i][j] &&
                        gradient[x][y]>gradient[i][j])))
                    {
                    i=x;
                    j=y;
                    }
        if (i==-1)
            break;
        gradient[i][j]=-1;
        dellabel(labelx[i][j],labely[i][j],level);
        }
    }


static void checklabel(float x,float y)

    {
    int     quadx,quady;
    float   px,py;
    float   grad;

    if (legend)
        return;
    quady=(y-yb1)*ldy/(.01+yb2-yb1);
    quadx=(x-xb1)*ldx/(.01+xb2-xb1);
    px=x_ps(x);
    py=y_ps(y);
    if (gradient[quadx][quady]<0)
        {
        if (!toocrowded(px,py,labeldx))
            {
            labelx[quadx][quady]=x;
            labely[quadx][quady]=y;
            gradient[quadx][quady]=calcgrad(x,y);
            addlabel(px,py,labeldx,level);
            }
        return;
        }
    grad=calcgrad(x,y);
    if (grad<gradient[quadx][quady] && !toocrowded(px,py,labeldx))
        {
        dellabel(labelx[quadx][quady],labely[quadx][quady],level);
        labelx[quadx][quady]=x;
        labely[quadx][quady]=y;
        gradient[quadx][quady]=grad;
        addlabel(px,py,labeldx,level);
        }
    }


static float calcgrad(float x,float y)

    {
    int     ix,iy;
    float   z1,z2,z3,dz1,dz2,dz3;

    ix=x;
    x=x-ix;
    if (ix>=xb2)
        {
        ix=xb2-1;
        x=.99;
        }
    iy=y;
    y=y-iy;
    if (iy>=yb2)
        {
        iy=yb2-1;
        y=.99;
        }
    if (x>y)
        z1=dp->shotdata[ix+1][iy];
    else
        z1=dp->shotdata[ix][iy+1];
    if (y>1-x)
        z2=dp->shotdata[ix+1][iy+1];
    else
        z2=dp->shotdata[ix][iy];
    z3=dp->interp[ix][iy];
    dz1=fabs(z1-z2);
    dz2=fabs(z2-z3);
    dz3=fabs(z3-z1);
    return(AMAX3(dz1,dz2,dz3));
    }


static void addlabel(float x,float y,float dx,float level)

    {
    int     i,j;
    LABLST *l;

    l=labels;
    for (i=0,j=0;j<labelcount;j++)
        {
        if (l->x1[i]<0)
            {
            l->x1[i]=x-dx;
            l->x2[i]=x+dx;
            l->y1[i]=y-fsize/2.;
            l->y2[i]=y+fsize/2.;
            l->level[i]=level;
            l->color[i]=ccolor;
            return;
            }
        if (i>=LABMAX-1)
            {
            l=l->next;
            i=0;
            }
        else
            i++;
        }
    i=labelcount;
    if (i)
        l=labels;
    while (i>LABMAX)
        {
        i-=LABMAX;
        l=l->next;
        }
    if (!i)
        {
        labels=malloc(sizeof(LABLST));
        labels->next=NULL;
        l=labels;
        }
    else
        if (i==LABMAX)
            {
            l->next=malloc(sizeof(LABLST));
            l=l->next;
            l->next=NULL;
            i=0;
            }
    l->x1[i]=x-dx;
    l->x2[i]=x+dx;
    l->y1[i]=y-fsize/2.;
    l->y2[i]=y+fsize/2.;
    l->level[i]=level;
    l->color[i]=ccolor;
    labelcount++;
    }


static void dellabel(float x,float y,float level)

    {
    float   px,py;
    int     i,j;
    LABLST *l;

    px=x_ps(x);
    py=y_ps(y);
    l=labels;
    for (i=0,j=0;j<labelcount;j++)
        {
        if (l->x1[i]>0 && l->level[i]==level && px>l->x1[i] && px<l->x2[i]
              && py>l->y1[i] && py<l->y2[i])
            {
            l->x1[i]=-1.;
            return;
            }
        if (i>=LABMAX-1)
            {
            l=l->next;
            i=0;
            }
        else
            i++;
        }
    printf("Internal error:  Unable to delete label.\n");
    }


static void printlabels(void)

    {
    float   px,py;
    int     xx,yy,dx,dy,i,j;
    char    buf[40];
    LABLST *l;

    if (postscript)
        {
        clearout();
        g_whitemar(1.05,1.);
        g_justify(4);
        }
    else
        {
        dy=4;
        setfillstyle(SOLID_FILL,1);
        settextjustify(CENTER_TEXT,CENTER_TEXT);
        }
    l=labels;
    for (i=0,j=0;j<labelcount;j++)
        {
        if (l->x1[i]>=0)
            {
            sprintf(buf,"%g",l->level[i]);
            px=(l->x1[i]+l->x2[i])/2.;
            py=(l->y1[i]+l->y2[i])/2.;
            if (postscript)
                {
                g_moveto(px,py);
                setgrey(l->color[i]);
                g_showtext(buf);
                }
            else
                {
                dx=strlen(buf)<<2;
                xx=vgax0+(px-psx1)*xscale_vga/xscale_ps;
                yy=vgay0+(py-psy1)*yscale_vga/yscale_ps;
                setcolor(l->color[i]);
                bar(xx-dx,yy-dy,xx+dx,yy+dy);
                outtextxy(xx,yy,buf);
                }
            }
        if (i>=LABMAX-1)
            {
            l=l->next;
            i=0;
            }
        else
            i++;
        }
    if (postscript)
        g_whitemar(0.,0.);
    }


static int toocrowded(float x,float y,float dx)

    {
    int     i,j;
    LABLST *l;

    if (x-dx<psx1 || x+dx>psx2 || y-fsize/2.<psy1 || y+fsize/2.>psy2)
        return(1);
    l=labels;
    for (i=0,j=0;j<labelcount;j++)
        {
        if (l->x1[i]>0 &&
             x+dx>=l->x1[i] && x-dx<=l->x2[i] && y+fsize/2.>=l->y1[i]
              && y-fsize/2.<=l->y2[i])
            return(1);
        if (i>=LABMAX-1)
            {
            l=l->next;
            i=0;
            }
        else
            i++;
        }
    return(0);
    }


static void freelabels(void)

    {
    while (labelcount)
        freelastlabel();
    }


static void freelastlabel(void)

    {
    int     i;
    int     newval;
    LABLST *l;
    LABLST *p;

    l=labels;
    p=NULL;
    i=labelcount;
    newval=0;
    while (i>LABMAX)
        {
        i-=LABMAX;
        newval+=LABMAX;
        p=l;
        l=l->next;
        }
    free(l);
    if (p==NULL)
        labels=NULL;
    else
        p->next=NULL;
    labelcount=newval;
    }
