/*
** CONTOUR.H    Master include file for all CONTOUR.C modules.
**
** Will Menninger, 11-12-92
**
*/


/*
** ANSI includes
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

/*
** Turbo C specific includes
*/
#include <conio.h>
#include <bios.h>
#include <mem.h>
#include <alloc.h>
#include <graphics.h>
#include <dos.h>

#define VERSION "2.38"


/*
** My library
*/
#include <willus4.h>

#define    NO_ERROR        1
#define    ERROR           0
#define    ESCAPE          -1
#define    ESC_KEY         0x1b
#define    MAXSTRLEN       80
#define    TRUE            1
#define    FALSE           0
#define    EOS             '\0'
#ifdef PI
#undef  PI
#endif
#define PI      3.1415926536

typedef int STATUS;
typedef int BOOLEAN;

typedef struct  {
                void            far*memory;
                float       far*far*shotdata;
                float       far*far*interp;
                float               max,min;
                int                 xsize;
                int                 ysize;
                long                stacksize;
                long                donesize;
                int            huge*stack;
                unsigned char   far*done;
                float               xmin,xmax,ymin,ymax;
                float               pagex,pagey,pageleftmar,pagebotmar;
                int                 no_title_margin;
                char                xtitle[100];
                char                ytitle[100];
                char                ztitle[100];
                char                gtitle[100];
                } DATA, *DATAPTR;


/*
** Function Prototypes
*/
/*--------- CONTOUR.C ----------------------------------------------*/
STATUS aborted (void);
void   debug   (char *s);

/*--------- LOAD.C -------------------------------------------------*/
BOOLEAN loaded     (BOOLEAN output);
void    prshotname (void);
float   xindex     (DATAPTR dp,float x);
float   yindex     (DATAPTR dp,float y);
float   dataset_min(DATAPTR dp,int x1,int y1,int x2,int y2);
float   dataset_max(DATAPTR dp,int x1,int y1,int x2,int y2);
STATUS  loadshot   (char *path,DATAPTR d);
STATUS  saveshot   (char *path,DATAPTR d,int status);

/*--------- IPLOT.C ------------------------------------------------*/
void iplot (DATAPTR d);

/*--------- 3D.C ---------------------------------------------------*/
void do_3d(DATAPTR d);
void clip_line(float x1,float y1,float z1,float x2,float y2,float z2,
               float zlo,float zhi,float *xa,float *ya,float *za,
               int *j);

/*--------- AREA.C -------------------------------------------------*/
void    init_avars(DATAPTR x);
void    area_stuff(void);

/*--------- AXES.C -------------------------------------------------*/
void    getticks (float x1,float x2,float *offset,float *pertick,
                  int *first,int *last,int *modval,int nticks);
void    init_ranges(DATAPTR d);
int     get_range (char axis,DATAPTR d,float *min,float *max);

/*--------- GRIDS.C ------------------------------------------------*/
void    setgriddim   (float x1,float y1,float x2,float y2,
                      int px1,int py1,int px2,int py2,
                      char *xf,char *yf);
void    bbox         (void);
/*
int     allocgridmem (void);
void    freegrid     (void);
void    getgriddata  (void);
void    putgriddata  (void);
*/
void    putgrid      (int grids);
void    put_titles   (DATAPTR d,int xt,int yt,int xx,int yx,int xy,int yy);

void init_grids(void);
void free_grids(void);
int  erase_grids(void);

/*--------- GRAPHIC.C ----------------------------------------------*/
BOOLEAN no_graphics  (void);
STATUS  graphics_on  (void);
void    graphics_off (void);
void    initpatterns (int color);
void    setupcolors  (int color,int dark_to_light);
BOOLEAN canswitch    (void);
void    rgbset       (int r,int g,int b,int cnum);
void    putpattern   (int x,int y,float f);
int     getpdy       (void);

/*--------- PSHELP.C -----------------------------------------------*/
int     nextname     (int color,char *s,int *append,int *eject);
int     fontsize     (float px,float py);
void ps_setgriddim   (float x1,float y1,float x2,float y2,
                      float px1,float py1,float px2,float py2,
                      float fsize,float axtck,float axtm,
                      char *xf,char *yf);
void ps_bbox         (void);
void ps_putgrid      (int grids);
void ps_put_titles   (DATAPTR d,float xt,float yt,float xx,float yx,
                      float xy,float yy);
void ps_update       (int prompt,int color,char *message,void *ptr);
void ps_clr_update   (void);
void ps_delete       (void);
void finalprint      (char *filename,int okay);

/*--------- PSCLINES.C ---------------------------------------------
void ps_ctrlines(char *filename,DATAPTR d,int x1,int y1,int x2,int y2,
                 char *ctrstring,int grids,int lx,int ly,int color);
*/

/*--------- PS3D.C -------------------------------------------------*/
void ps_3d(char *filename,DATAPTR dp,int xo,int xs,int yo,int ys,
           float zo,float zs,int vt,int grids,int color,
           float *zb,int nl,int sl,int _append,int _eject);

/*--------- PSIPLOT.C ----------------------------------------------*/
void ps_iplot        (char *filename,DATAPTR d,float zmin,float zmax,
                      int grids,int color,int dark_to_light,int append,
                      int eject);

/*--------- CLPLOT.C -----------------------------------------------*/
void  contour_plot   (DATAPTR d);
void  putpoint       (int x,int y,int v);
void  filltri        (int ix,int iy,int nt);
void  curve_start    (float x,float y,int closed,int pc);
void  curve_next     (float x,float y);
void  curve_end      (void);
int   need_count     (void);
void  drawline       (float x1,float y1,float x2,float y2);
void  contour_colors (int color);

/*--------- CLCALC.C -----------------------------------------------*/
int   one_level      (DATAPTR d,int x1,int y1,int x2,int y2,float level);


/*--------- SAVEPCX.C ----------------------------------------------*/
char *vgasave        (void);

/*--------- BATCH.C ------------------------------------------------*/
void setbatch(char *filename);
void killbatch(void);
int  bmenu(char *m[]);
int  bbioskey(int arg);
int  bgetinput(char *prompt,char *mbuf,int maxlen);
