/*
**
** GRAPHIC.C   Handles IBM specific graphics operations and provides
**	       a set of higher level routines.
**
** Will Menninger, 9/89, programmed in Turbo C 2.0
**
*/

#include "contour.h"

static int	       gmode;
static int	       gdriver;
static unsigned char   patterns[3300];
static int	       pdy;
static int	       psize;
static int	       pmax;

static void writevgapal (int cnum,int r,int g,int b);



BOOLEAN no_graphics(void)

    {
    detectgraph((int far *)&gdriver,(int far *)&gmode);
    if ((gdriver!=EGA || gmode!=EGAHI) && (gdriver!=VGA || gmode!=VGAHI))
	{
	printf("An EGA or a VGA graphics adapter is required to run "
	       "this software.\n");
	return(TRUE);
	}
    registerbgidriver(EGAVGA_driver);
    return(FALSE);
    }


STATUS graphics_on(void)

    {
    int status;

    initgraph((int far *)&gdriver,(int far *)&gmode,(char far *)"");
    status=graphresult();
    if (status<0)
	{
	error("");
	cprintf("Graphics error:  %s!",grapherrormsg(status));
	normal("\r\n\r\n");
	wait_enter();
	return(ERROR);
	}
    return(NO_ERROR);
    }


void graphics_off(void)

    {
    closegraph();
    }


void initpatterns(int color)

    {
    static int inited=0;
    static int icolor;
    static int egapats[5]={0,1,0x41,0x45,0xa5};
    static int vgapats[7]={0,1,0x41,0x141,0x4a1,0x585,0x5a5};
    char       temp[30];
    int        s;
    int        cmax,rep;
    int        i,j,x,y,mask;

    if (inited && (gdriver==VGA || icolor==color))
	return;
    getimage(0,0,3,2,temp);
    if (gdriver==EGA)
	{
	psize=imagesize(0,0,3,1);
	pdy=2;
	cmax=color ? 14 : 3;
	rep=8;
	pmax=rep*cmax+1;
	}
    else
	{
	psize=imagesize(0,0,3,2);
	pdy=3;
	cmax=14;
	rep=12;
	pmax=rep*cmax+1;
	}
    s=0;
    for (i=1;i<=cmax;i++)
	for (j=0;j<rep;j++)
	    {
	    if (gdriver==EGA)
		mask=j>rep/2 ? ~egapats[rep-j] : egapats[j];
	    else
		mask=j>rep/2 ? ~vgapats[rep-j] : vgapats[j];
	    for (y=0;y<pdy;y++)
		for (x=0;x<4;x++,mask>>=1)
		    putpixel(x,y,mask&1 ? i+1 : i);
	    getimage(0,0,3,pdy-1,&patterns[s]);
	    s+=psize;
	    }
    for (y=0;y<pdy;y++)
	for (x=0;x<4;x++)
	    putpixel(x,y,i);
    getimage(0,0,3,pdy-1,&patterns[s]);
    putimage(0,0,temp,COPY_PUT);
    inited=1;
    icolor=color;
    }


int  getpdy(void)

    {
    return(pdy);
    }


void putpattern(int x,int y,float f)

    {
    int     n;

    n=(f==1 ? pmax-1 : (f-floor(f))*pmax);
    putimage(x,y,&patterns[n*psize],COPY_PUT);
    }


void setupcolors(int color,int dark_to_light)

    {
    static int d2l[7][4]={{80,80,80,6},{0,0,0,1},{0,0,255,2},{255,0,0,3},
			  {0,255,0,3},{255,255,0,3},{255,255,255,3}};
    static int l2d[7][4]={{80,80,80,6},{255,255,255,1},{255,255,0,3},
			  {0,255,0,3},{255,0,0,3},{0,0,255,3},{0,0,0,2}};
    int        cmap[7][4];
    int        i,j,n,r,g,b;

    if (dark_to_light)
	for (i=0;i<=d2l[0][3];i++)
	    for (j=0;j<4;j++)
		cmap[i][j]=d2l[i][j];
    else
	for (i=0;i<=l2d[0][3];i++)
	    for (j=0;j<4;j++)
		cmap[i][j]=l2d[i][j];

    if (!color)
	{
	rgbset(0,0,255,0);
	n= gdriver==EGA ? 4 : 15;
	for (i=0;i<n;i++)
	    {
	    j=i*255/(n-1);
	    if (!dark_to_light)
		j=255-j;
	    rgbset(j,j,j,i+1);
	    }
	}
    else
	{
	rgbset(cmap[0][0],cmap[0][1],cmap[0][2],0);
	n=1;
	for (i=1;i<=cmap[0][3];i++)
	    for (j=1;j<=cmap[i][3];j++)
		{
		r=cmap[i-1][0]+(cmap[i][0]-cmap[i-1][0])*j/cmap[i][3];
		g=cmap[i-1][1]+(cmap[i][1]-cmap[i-1][1])*j/cmap[i][3];
		b=cmap[i-1][2]+(cmap[i][2]-cmap[i-1][2])*j/cmap[i][3];
		rgbset(r,g,b,n++);
		}
	}
    }


void rgbset(int r,int g,int b,int cnum)

    {
    int egapal;

    if (gdriver==VGA)
	{
	writevgapal(cnum,r>>2,g>>2,b>>2);
	return;
	}
    egapal= ((r&0x80)>>5) | ((r&0x40)>>1) | ((g&0x80)>>6) |
	    ((g&0x40)>>2) | ((b&0x80)>>7) | ((b&0x40)>>3);
    setpalette(cnum,egapal);
    }


static void writevgapal(int cnum,int r,int g,int b)

    {
    union REGS	    s;

    s.h.bl=cnum;
    s.x.ax=0x1007;
    int86(0x10,&s,&s);
    s.h.bl=s.h.bh;
    s.h.bh=0;
    s.x.ax=0x1010;
    s.h.dh=r;
    s.h.ch=g;
    s.h.cl=b;
    int86(0x10,&s,&s);
    }


BOOLEAN canswitch(void)

    {
    return(gdriver==VGA);
    }
