/*
**
** GRIDS.C      Allows removable grids if enough memory exists
**
** Will Menninger, 11-19-92
**
*/

#include "contour.h"

#define GRIDCOLOR   0

static float    gr_x1,gr_y1,gr_x2,gr_y2;
static int      px_x1,px_y1,px_x2,px_y2;
static float    yoff,yscale,xoff,xscale;
static int      ix1,iy1,ix2,iy2;
static int      xmod,ymod;
static float    yppu,xppu;
static char     xaxformat[20];
static char     yaxformat[20];

struct gridstruct {
       int      x1;
       int      y1;
       int      x2;
       int      y2;
       int      memsize;
       char far *data;
       struct gridstruct *prev;
       struct gridstruct *next;
       };

typedef struct gridstruct GRIDLINE;

static GRIDLINE grid0;
static void store_grid(int x1,int y1,int x2,int y2,GRIDLINE **g,int *store_it);


/*
** (x1,y1) = lower left corner, (x2,y2)=upper right corner
** (px1,py1) = lower left pixel corner
** (px2,py2) = upper right pixel corner (py2 < py1)!
**
** The pixel coordinates define the outlining box.  They are
** NOT included in the graph.  I.e. the graph area goes from
** (px1+1,py1-1) (lower left) to (px2-1,py2+1) (upper right)
**
*/
void setgriddim(float x1,float y1,float x2,float y2,
                int px1,int py1,int px2,int py2,
                char *xf,char *yf)

    {
    gr_x1=x1;
    gr_y1=y1;
    gr_x2=x2;
    gr_y2=y2;
    px_x1=px1;
    px_x2=px2;
    px_y1=py1;
    px_y2=py2;
    strcpy(xaxformat,xf);
    strcpy(yaxformat,yf);
    yppu=(float)(px_y1-px_y2)/(gr_y2-gr_y1);
    xppu=(float)(px_x2-px_x1)/(gr_x2-gr_x1);
    getticks(gr_x1,gr_x2,&xoff,&xscale,&ix1,&ix2,&xmod,15);
    getticks(gr_y1,gr_y2,&yoff,&yscale,&iy1,&iy2,&ymod,15);
    }


void init_grids(void)

    {
    grid0.prev=NULL;
    grid0.next=NULL;
    grid0.data=NULL;
    }


void free_grids(void)

    {
    GRIDLINE  *temp;

    for (temp=&grid0;temp->next!=NULL;temp=temp->next);
    while (1)
        {
        if (temp->data!=NULL)
            farfree(temp->data);
        temp->data=NULL;
        if (temp->prev==NULL)
            break;
        temp=temp->prev;
        free(temp->next);
        temp->next=NULL;
        }
    }


void bbox(void)

    {
    setcolor(15);
    rectangle(px_x1,px_y1,px_x2,px_y2);
    rectangle(px_x1-1,px_y1+1,px_x2+1,px_y2-1);
    rectangle(px_x1-2,px_y1+1,px_x2+2,px_y2-1);
    }


void putgrid(int grids)

    {
    int     store_it;
    int     i,pxval;
    float   xval;
    char    buf[100];
    GRIDLINE *gridline;

    store_it=1;
    if (grids && grid0.data!=NULL)
        free_grids();
    gridline=&grid0;
    setcolor(GRIDCOLOR);
    settextjustify(RIGHT_TEXT,CENTER_TEXT);
    for (i=iy1;i<=iy2;i++)
        {
        xval=yoff+i*yscale;
        if (fabs(xval/yscale)<1e-4)
            xval=0;
        pxval=(float)px_y1+.5-(xval-gr_y1)*yppu;
        if (pxval>=px_y1 || pxval<=px_y2)
            continue;
        if (grids)
            {
            store_grid(px_x1+1,pxval,px_x2-1,pxval,&gridline,&store_it);
            line(px_x1+1,pxval,px_x2-1,pxval);
            }
        if (i%ymod==0)
            {
            sprintf(buf,yaxformat,xval);
            setcolor(13);
            outtextxy(px_x1-7,pxval,buf);
            setcolor(GRIDCOLOR);
            if (grids)
                if (pxval==px_y1-1)
                    {
                    store_grid(px_x1+1,pxval-1,px_x2-1,pxval-1,&gridline,&store_it);
                    line(px_x1+1,pxval-1,px_x2-1,pxval-1);
                    }
                else
                    {
                    store_grid(px_x1+1,pxval+1,px_x2-1,pxval+1,&gridline,&store_it);
                    line(px_x1+1,pxval+1,px_x2-1,pxval+1);
                    }
            }
        }
    settextjustify(CENTER_TEXT,TOP_TEXT);
    for (i=ix1;i<=ix2;i++)
        {
        xval=i*xscale+xoff;
        if (fabs(xval/xscale)<1e-4)
            xval=0;
        pxval=(float)px_x1+.5+(xval-gr_x1)*xppu;
        if (pxval<=px_x1 || pxval>=px_x2)
            continue;
        if (grids)
            {
            store_grid(pxval,px_y2+1,pxval,px_y1-1,&gridline,&store_it);
            line(pxval,px_y2+1,pxval,px_y1-1);
            }
        if (i%xmod==0)
            {
            sprintf(buf,xaxformat,xval);
            setcolor(13);
            outtextxy(pxval,px_y1+4,buf);
            setcolor(GRIDCOLOR);
            if (grids)
                if (pxval==px_x2-1)
                    {
                    store_grid(pxval-1,px_y2+1,pxval-1,px_y1-1,&gridline,&store_it);
                    line(pxval-1,px_y2+1,pxval-1,px_y1-1);
                    }
                else
                    {
                    store_grid(pxval+1,px_y2+1,pxval+1,px_y1-1,&gridline,&store_it);
                    line(pxval+1,px_y2+1,pxval+1,px_y1-1);
                    }
            }
        }
    }


static void store_grid(int x1,int y1,int x2,int y2,GRIDLINE **g,int *store_it)

    {
    GRIDLINE *gptr;
    int     temp;
    int     memsize;
    long    memcheck;

    if (!(*store_it))
        return;
    if (x1>x2)
        {
        temp=x2;
        x2=x1;
        x1=temp;
        }
    if (y1>y2)
        {
        temp=y2;
        y2=y1;
        y1=temp;
        }
    (*g)->memsize=imagesize(x1,y1,x2,y2);
    memsize=(*g)->memsize+4;
    memcheck=farcoreleft();
    (*g)->data=farmalloc((unsigned long)memsize);
    if ((*g)->data==NULL || memcheck==farcoreleft())
        {
        free_grids();
        (*store_it)=0;
        return;
        }
    getimage(x1,y1,x2,y2,(*g)->data);
    (*g)->x1=x1;
    (*g)->y1=y1;
    (*g)->x2=x2;
    (*g)->y2=y2;
    (*g)->next=malloc(sizeof(GRIDLINE));
    if ((*g)->next==NULL)
        {
        free_grids();
        (*store_it)=0;
        return;
        }
    gptr=(*g);
    (*g)=(*g)->next;
    (*g)->prev=gptr;
    (*g)->data=NULL;
    (*g)->next=NULL;
    }


int erase_grids(void)

    {
    GRIDLINE *temp;

    if (grid0.data==NULL)
        return(0);
    for (temp=&grid0;temp->next!=NULL;temp=temp->next);
    while (1)
        {
        if (temp->data!=NULL)
            putimage(temp->x1,temp->y1,temp->data,COPY_PUT);
        if (temp->prev==NULL)
            break;
        temp=temp->prev;
        }
    return(1);
    }


void put_titles(DATAPTR d,int xt,int yt,int xx,int yx,int xy,int yy)

    {
    settextjustify(CENTER_TEXT,TOP_TEXT);
    outtextxy(xt,yt,d->gtitle);
    outtextxy(xx,yx+14,d->xtitle);
    settextstyle(0,VERT_DIR,0);
    settextjustify(RIGHT_TEXT,CENTER_TEXT);
    outtextxy(xy-60,yy,d->ytitle);
    settextstyle(0,HORIZ_DIR,0);
    }
