/*
**
** IPLOT.C     Does an intensity plot of the CCD picture
**
** Will Menninger, 9/89, programmed in Turbo C 2.0
**
** Modified 6-1-92 to deal with a general rectangular grid.
**
*/

#include "contour.h"

#define XL1     57
#define XL2     70
#define XB      140
#define BOXH    120
#define BOXW    480

static float    zmin;
static float    zscale;


void iplot(DATAPTR d)

    {
    int             i,in,ix,iy,x,y,x0,y0,grids_on;
    int             i1,i2,modval,not_clear;
    float           offset,scale;
    int             pdy,px,py,maxx,maxy;
    char            buf[100];
    char            prompt[100];
    float           dmin,z,zmax;
    int             ss,color,dark_to_light;

    if (!loaded(TRUE))
        return;
    if (!canswitch())
        {
        cursoroff();
        normal("Press <C> for color, <G> for grey-scale.");
        while (1)
            {
            i=bbioskey(0)&0xff;
            if (i==ESC_KEY)
                return;
            if (i=='c' || i=='C' || i=='g' || i=='G')
                break;
            }
        if (i!=ESC_KEY)
            {
            cputs("\r\n\r\n");
            color=(i=='c' || i=='C');
            }
        }
    if (!get_range('Z',d,&zmin,&zmax))
        return;
    /*
    if (d->min>=0.)
        dmin=0.;
    else
        dmin=d->min;
    */
    dmin=d->min;
    zscale=zmax-zmin;
    /*
    cursoroff();
    normal("Go from Dark to Light in color (Y or N) [Def=Y] ?");
    while (1)
        {
        i=bbioskey(0)&0xff;
        if (i==ESC_KEY)
            return;
        if (i=='y' || i=='Y' || i=='n' || i=='N' || i==13)
            break;
        }
    cputs("\r\n\r\n");
    dark_to_light=(i=='y' || i=='Y' || i==13);
    */
    dark_to_light=1;
    if (graphics_on()!=NO_ERROR)
        return;
    initpatterns(color);
    setupcolors(color,dark_to_light);
    pdy=getpdy();
    x0=XB;
    y0=(getmaxy()-pdy*BOXH)/2;
    setcolor(15);

    /* Legend colors */
    if (zmin==dmin && zscale==d->max-dmin)
        ss=pdy;
    else
        ss=1;
    for (y=0;y<BOXH*pdy;y+=ss)
        {
        z=dmin+(d->max-dmin)*y/(pdy*BOXH-1.);
        for (x=0;x<3;x++)
            putpattern(XL1+1+x*4,y0+y,(z-zmin)/zscale);
        }
    setfillstyle(SOLID_FILL,0);
    bar(XL1+1,y0+BOXH*pdy,XL1+21,y0+BOXH*pdy+pdy+2);

    settextjustify(RIGHT_TEXT,CENTER_TEXT);

    /* Legend rectangle */
    rectangle(XL1,y0-1,XL2,y0+pdy*BOXH);

    /* Legend values */
    getticks(dmin,d->max,&offset,&scale,&i1,&i2,&modval,25);
    for (i=i1;i<=i2;i++)
        {
        if (i%modval)
            continue;
        z=offset+i*scale;
        if (fabs(z/scale)<1e-4)
            z=0;
        sprintf(buf,"%g",z);
        y=y0-1+(z-dmin)*(pdy*BOXH+1)/(d->max-dmin);
        outtextxy(XL1-7,y,buf);
        line(XL1-4,y,XL1,y);
        }

    /* Titles */

    put_titles(d,320,15,x0+BOXW/2.,y0+pdy*BOXH,XL2+80,y0+pdy*BOXH/2.);
    settextjustify(LEFT_TEXT,BOTTOM_TEXT);
    outtextxy(0,y0-8,d->ztitle);

    /* Box colors */
    if (d->ysize>BOXH*pdy)
        maxy=BOXH*pdy;
    else
        if (d->ysize<BOXH)
            maxy=BOXH;
        else
            maxy=d->ysize;
    if (d->xsize>BOXW)
        maxx=BOXW;
    else
        if (d->xsize<(BOXW>>2))
            maxx=BOXW>>2;
        else
            maxx=d->xsize;
    for (y=0;y<maxy;y++)
        {
        if (bbioskey(1))
            {
            bbioskey(0);
            break;
            }
        iy=(float)(maxy-1-y)*(float)d->ysize/(float)maxy;
        py=y0+(float)y*pdy*BOXH/(float)maxy;
        for (x=0;x<maxx;x++)
            {
            ix=(float)x*(float)d->xsize/(float)maxx;
            px=x0+(float)x*BOXW/(float)maxx;
            putpattern(px,py,(d->shotdata[ix][iy]-zmin)/zscale);
            }
        }

    setgriddim(d->xmin-.5*(d->xmax-d->xmin)/(d->xsize-1.),
               d->ymin-.5*(d->ymax-d->ymin)/(d->ysize-1.),
               d->xmax+.5*(d->xmax-d->xmin)/(d->xsize-1.),
               d->ymax+.5*(d->ymax-d->ymin)/(d->ysize-1.),
               x0-1,y0+pdy*BOXH,x0+BOXW,y0-1,"%g","%g");

    /* Bounding rectangle */
    bbox();

    /* Toggle between color and grey-scale if VGA */
    grids_on=0;
    putgrid(0);
    not_clear=0;
    strcpy(prompt,"<T> color/bw, <G> grid, <P> print, <S> save, "
                  "<I> inverse, <ESC> exit");
    ps_update(0,10,prompt,NULL);
    while (1)
        {
        in=bbioskey(0)&0xff;
        if (not_clear)
            {
            not_clear=0;
            ps_clr_update();
            }
        if (in=='s' || in=='S')
            {
            ps_update(0,15,vgasave(),NULL);
            not_clear=1;
            continue;
            }
        if (in=='p' || in=='P')
            {
            int append,eject;

            if (nextname(10,buf,&append,&eject))
                {
                ps_iplot(buf,d,zmin,zmax,!i,color,dark_to_light,append,eject);
                not_clear=1;
                }
            else
                ps_update(0,10,prompt,NULL);
            continue;
            }
        if (canswitch() && (in=='t' || in=='T'))
            {
            color=!color;
            setupcolors(color,dark_to_light);
            continue;
            }
        if (in=='i' || in=='I')
            {
            dark_to_light=!dark_to_light;
            setupcolors(color,dark_to_light);
            continue;
            }
        if (in=='g' || in=='G')
            {
            if (grids_on && erase_grids())
                grids_on=0;
            else
                {
                putgrid(1);
                grids_on=1;
                }
            continue;
            }
        break;
        }
    free_grids();
    graphics_off();
    }
