/*
**
** LOAD.C      Handles disk I/O for CONTOUR.C
**
** Will Menninger, 9/89, programmed in Turbo C 2.0
**
** Modified 6-1-92 for any size rectangular grid.
**
*/

#include "contour.h"

#define MAXGRAPHS   100

static BOOLEAN shot_loaded=0;
static char shotname[80];
static long graphptr[MAXGRAPHS];
static int  ngraphs=0;

static int numgraphs(FILE *f);
static int goto_next_graph(FILE *f,int quiet);
static void shotsize(FILE *f,int *x,int *y,int *bin,long maxpos);
static BOOLEAN alloc_mem(DATAPTR x);
static int setting(DATAPTR dp,FILE *s);
static STATUS premature(FILE *s);
static void sample_save(FILE *s);


BOOLEAN loaded(BOOLEAN output)

    {
    if (output && !shot_loaded)
        {
        error("There is no data in memory!");
        normal("\r\n\r\n");
        wait_enter();
        }
    return(shot_loaded);
    }


void prshotname(void)

    {
    if (shot_loaded)
        hilite(shotname);
    }


float xindex(DATAPTR dp,float x)

    {
    if (dp->xmax==dp->xmin)
        return(0);
    return((x-dp->xmin)*(dp->xsize-1.)/(dp->xmax-dp->xmin));
    }


float yindex(DATAPTR dp,float y)

    {
    if (dp->ymax==dp->ymin)
        return(0);
    return((y-dp->ymin)*(dp->ysize-1.)/(dp->ymax-dp->ymin));
    }


float dataset_min(DATAPTR dp,int x1,int y1,int x2,int y2)

    {
    int     x,y;
    float   min;

    min=dp->shotdata[x1][y1];
    for (x=x1;x<=x2;x++)
        for (y=y1;y<=y2;y++)
            if (min>dp->shotdata[x][y])
                min=dp->shotdata[x][y];
    return(min);
    }


float dataset_max(DATAPTR dp,int x1,int y1,int x2,int y2)

    {
    int     x,y;
    float   max;

    max=dp->shotdata[x1][y1];
    for (x=x1;x<=x2;x++)
        for (y=y1;y<=y2;y++)
            if (max<dp->shotdata[x][y])
                max=dp->shotdata[x][y];
    return(max);
    }



/*
** status = 0:  Save data
** status = 1:  Save rotated data (90 deg)
** status = 2:  Save sample data
*/
STATUS saveshot(char *path,DATAPTR d,int status)

    {
    char    filename[80];
    FILE   *s;
    int     i,j,c;

    if (status!=2 && !loaded(1))
        return(ERROR);
    while (1)
        {
        if (bgetinput("Save name:  ",filename,60)==ESCAPE)
            return(ESCAPE);
        normal("\r\n\r\n");
        if (filename[0]==EOS)
            return(ESCAPE);
        defaultpath(filename,path);
        defaultext(filename,".DAT");
        strupr(filename);
        s=fopen(filename,"rb");
        if (s!=NULL)
            {
            fclose(s);
            error("");
            cprintf("The file %s already exists!  Overwrite?  <Y> or <N>",
                    filename);
            normal("\r\n\r\n");
            while (1)
                {
                c=bbioskey(0)&0xff;
                if (c==27)
                    return(ESCAPE);
                c=tolower(c);
                if (c=='y' || c=='n')
                    break;
                }
            if (c=='n')
                continue;
            }
        break;
        }
    s=fopen(filename,"w");
    if (s==NULL)
        {
        error("");
        cprintf("Cannot open file %s for writing!\n",filename);
        normal("\r\n\r\n");
        wait_enter();
        return(ERROR);
        }
    cprintf("Writing to %s ",filename);
    fprintf(s,";\n; Data save from CONTOUR " VERSION "\n;\n");
    if (status==2)
        sample_save(s);
    else if (status==0)
        {
        if (d->pagex>0 && d->pagey>0)
            fprintf(s,"/fpo pd %g %g\n",d->pagex,d->pagey);
        if (d->gtitle[0]!=EOS)
            fprintf(s,"/et g \"%s\"\n",d->gtitle);
        if (d->xtitle[0]!=EOS)
            fprintf(s,"/et x \"%s\"\n",d->xtitle);
        if (d->ytitle[0]!=EOS)
            fprintf(s,"/et y \"%s\"\n",d->ytitle);
        if (d->ztitle[0]!=EOS)
            fprintf(s,"/et z \"%s\"\n",d->ztitle);
        fprintf(s,"//cx %g %g\n",d->xmin,d->xmax);
        fprintf(s,"//cy %g %g\n",d->ymin,d->ymax);
        fprintf(s,"/td z\n");
        for (j=d->ysize-1;j>=0;j--)
            {
            for (i=0;i<d->xsize-1;i++)
                fprintf(s,"%g ",d->shotdata[i][j]);
            fprintf(s,"%g\n",d->shotdata[d->xsize-1][j]);
            if ((j%10)==0)
                cprintf(".");
            }
        }
    else
        {
        if (d->pagex>0 && d->pagey>0)
            fprintf(s,"/fpo pd %g %g\n",d->pagex,d->pagey);
        if (d->gtitle[0]!=EOS)
            fprintf(s,"/et g \"%s\"\n",d->gtitle);
        if (d->xtitle[0]!=EOS)
            fprintf(s,"/et x \"%s\"\n",d->ytitle);
        if (d->ytitle[0]!=EOS)
            fprintf(s,"/et y \"%s\"\n",d->xtitle);
        if (d->ztitle[0]!=EOS)
            fprintf(s,"/et z \"%s\"\n",d->ztitle);
        fprintf(s,"//cx %g %g\n",d->ymin,d->ymax);
        fprintf(s,"//cy %g %g\n",d->xmin,d->xmax);
        fprintf(s,"/td z\n");
        for (i=d->xsize-1;i>=0;i--)
            {
            for (j=0;j<=d->ysize-1;j++)
                fprintf(s,"%g%s",d->shotdata[i][j],j<d->ysize-1 ? " " : "\n");
            if ((i%10)==0)
                cprintf(".");
            }
        }
    fclose(s);
    return(NO_ERROR);
    }


static void sample_save(FILE *s)

    {
    int     i,j;
    float   x,y,z,r;

    fprintf(s,"/fpo pd 6 6\n/et g \"Delicious Doughnut\"\n"
              "/et x \"X-dimension (in)\"\n"
              "/et y \"Y-dimension (cm)\"\n"
              "/et z \"Height (in)\"\n"
              "//cx -3 3\n//cy -7.62 7.62\n");
    for (j=50;j>=0;j--)
        for (i=0;i<=50;i++)
            {
            x=3.*(i-25)/25.;
            y=3.*(j-25)/25.;
            r=sqrt(x*x+y*y);
            if (r<=1 || r>=2)
                z=0.;
            else
                z=sqrt(.25-(1.5-r)*(1.5-r));
            fprintf(s,"%4.3f%s",z,i<50 ? " " : "\n");
            }
    }


STATUS loadshot(char *path,DATAPTR d)

    {
    char    prompt[80];
    char    buf[80];
    char    numbuf[80];
    int     y,x,ix,iy,i,c,imod;
    int     selected,binary;
    float   f,max,min;
    int     newline;
    int     xstep,ystep;
    int     ng,orgxsize,orgysize;
    long    maxpos,startpos,pos;
    FILE   *s;

    while (1)
        {
        sprintf(prompt,"Data file [%s]:  ",shot_loaded ? shotname : ".DAT");
        if (bgetinput(prompt,buf,48)==ESCAPE)
            return(ESCAPE);
        normal("\r\n\r\n");
        if (buf[0]==EOS && shot_loaded)
            strcpy(buf,shotname);
        else
            {
            defaultpath(buf,path);
            defaultext(buf,".DAT");
            strupr(buf);
            }
        s=fopen(buf,"rb");
        if (s==NULL)
            {
            error("");
            cprintf("Can't open %s for reading!",buf);
            normal("\r\n\r\n");
            continue;
            }
        break;
        }
    /*
    if (!strcmp(shotname,buf))
        ng=ngraphs;
    else
        {
    */
        normal("Listing available plots ... \r\n");
        ng=numgraphs(s);
    /*
        }
    */
    /*
    normal("\r\n");
    sprintf(numbuf,"There are %d graph%s in %s.",ng,ng>1 ? "s" : "",buf);
    hilite(numbuf);
    */
    normal("\r\n");
    if (ng>1)
        {
        while (1)
            {
            if (bgetinput("Enter graph number [1]:  ",numbuf,6)==ESCAPE)
                return(ESCAPE);
            normal("\r\n");
            i= numbuf[0]==EOS ? 1 : atoi(numbuf);
            if (i<1 || i>ng)
                {
                error("");
                cprintf("There are %d graphs.  Choose between 1 and %d.\n",ng,ng);
                normal("\r\n");
                continue;
                }
            break;
            }
        selected=i;
        }
    normal("Analyzing ");
    hilite(buf);
    normal(" . . . ");
    fseek(s,0L,0);
    if (ng==1)
        {
        startpos=0L;
        fseek(s,0L,2);
        maxpos=ftell(s);
        fseek(s,0L,0);
        }
    else
        {
        if (selected>MAXGRAPHS)
            {
            fseek(s,graphptr[MAXGRAPHS-1],0);
            for (i=0;i<selected-MAXGRAPHS;i++)
                goto_next_graph(s,1);
            }
        else
            fseek(s,graphptr[selected-1],0);
        startpos=ftell(s);
        i=goto_next_graph(s,1);
        maxpos=ftell(s)-i;
        fseek(s,startpos,0);
        }
    shotsize(s,&(d->xsize),&(d->ysize),&binary,maxpos);
    if (d->xsize<0 ||  d->ysize<0)
        {
        wait_enter();
        return(ERROR);
        }
    normal("Load the data as is?  (Y/N)");
    while (1)
        {
        c=tolower(bbioskey(0)&0xff);
        if (c==27 || c==13 || c=='y' || c=='n')
            break;
        }
    if (c==27)
        return(ESCAPE);
    normal("\r\n");
    if (c=='n')
        {
        if (bgetinput("Enter Xstep,Ystep [Def=1,1]:  ",numbuf,20)==ESCAPE)
            return(ESCAPE);
        normal("\r\n");
        if (numbuf[0]!=EOS)
            c=sscanf(numbuf,"%d,%d",&xstep,&ystep);
        else
            c=0;
        if (!c)
            xstep=ystep=1;
        else
            if (c==1)
                ystep=xstep;
        }
    else
        {
        xstep=1;
        ystep=1;
        }
    orgxsize=d->xsize;
    orgysize=d->ysize;
    if (xstep>1)
        d->xsize=(d->xsize+xstep-1)/xstep;
    if (ystep>1)
        d->ysize=(d->ysize+ystep-1)/ystep;
    if (d->xsize==-1)
        {
        normal("\r\n\r\n");
        wait_enter();
        return(ERROR);
        }
    if (!alloc_mem(d))
        {
        normal("\r\n\r\n");
        wait_enter();
        return(ERROR);
        }
    normal("Loading ");
    hilite(buf);
    normal(" . . . ");
    d->xmin=0;
    d->xmax=orgxsize-1.;
    d->ymin=0;
    d->ymax=orgysize-1.;
    d->gtitle[0]=EOS;
    d->xtitle[0]=EOS;
    d->ytitle[0]=EOS;
    d->ztitle[0]=EOS;
    if (binary && (long)(orgxsize)*(long)(orgysize)!=maxpos-startpos)
        startpos+=4;
    fseek(s,startpos,0);
    pos=startpos;
    if (binary)
        for (iy=orgysize-1,y=d->ysize-1;iy>=0;iy--)
            {
            imod=iy%ystep;
            for (ix=0,x=0;ix<orgxsize;ix++)
                {
                c=fgetc(s);
                pos++;
                if (pos>maxpos || c==EOF)
                    return(premature(s));
                if (imod || ix%xstep)
                    continue;
                d->shotdata[x][y]=c;
                if (y==d->ysize-1 && x==0)
                    max=min=c;
                else
                    if (c>max)
                        max=c;
                    else
                        if (c<min)
                            min=c;
                x++;
                }
            if (!imod)
                y--;
            }
    else
        {
        newline=1;
        for (iy=orgysize-1,y=d->ysize-1;iy>=0;iy--)
            {
            imod=iy%ystep;
            for (ix=0,x=0;ix<orgxsize;ix++)
                {
                while (1)
                    {
                    while (((c=fgetc(s)),++pos,c==' ') || c=='\t' || c=='\r');
                    if (c=='\n')
                        {
                        newline=1;
                        continue;
                        }
                    if (pos>maxpos || c==EOF)
                        return(premature(s));
                    if (c==';')
                        {
                        while (pos<maxpos && ((c=fgetc(s)),++pos,c!='\n')
                                          && c!=EOF);
                        if (c=='\n')
                            {
                            newline=1;
                            continue;
                            }
                        return(premature(s));
                        }
                    if (newline && c=='/')
                        {
                        if (!setting(d,s))
                            return(premature(s));
                        pos=ftell(s);
                        continue;
                        }
                    newline=0;
                    numbuf[0]=c;
                    for (i=1;i<80 && ((c=fgetc(s)),++pos,c=='e') || c=='E' ||
                         c=='d' || c=='D' || (c>='0' && c<='9') || c=='.' ||
                         c=='-' || c=='+';i++)
                        numbuf[i]=c;
                    numbuf[i]=EOS;
                    if (c=='\n')
                        newline=1;
                    if (imod || ix%xstep)
                        break;
                    f=atof(numbuf);
                    d->shotdata[x][y]=f;
                    if (x==0 && y==d->ysize-1)
                        max=min=f;
                    else
                        if (f>max)
                            max=f;
                        else
                            if (f<min)
                                min=f;
                    x++;
                    break;
                    }
                }
            if (!imod)
                y--;
            }
        }
    fclose(s);

    /* Adjust for step sizes */

    if (orgxsize>1 && (orgxsize+xstep-1)%xstep)
        {
        i=d->xsize*xstep-(xstep-1);
        d->xmax=d->xmin+(d->xmax-d->xmin)*(float)(i-1)/(orgxsize-1);
        }
    if (orgysize>1 && (orgysize+ystep-1)%ystep)
        {
        i=d->ysize*ystep-(ystep-1);
        d->ymax=d->ymin+(d->ymax-d->ymin)*(float)(i-1)/(orgysize-1);
        }

    d->min=min;
    d->max=max;
    shot_loaded=TRUE;
    strcpy(shotname,buf);
    normal("Interpolating . . .");
    for (y=0;y<d->ysize-1;y++)
        for (x=0;x<d->xsize-1;x++)
            d->interp[x][y]=(d->shotdata[x][y]+d->shotdata[x+1][y]+
                          d->shotdata[x][y+1]+d->shotdata[x+1][y+1])/4;
    init_avars(d);
/*
    for (y=0;y<d->ysize;y++)
        {
        for (x=0;x<d->xsize;x++)
            cprintf("%5.2f ",d->shotdata[x][y]);
        cprintf("\r\n");
        }
    cprintf("\r\n");
    for (y=0;y<d->ysize-1;y++)
        {
        for (x=0;x<d->xsize-1;x++)
            cprintf("%5.2f ",d->interp[x][y]);
        cprintf("\r\n");
        }
*/
    return(NO_ERROR);
    }


static int setting(DATAPTR dp,FILE *s)

    {
    char    buf[150];
    char    buf2[150];
    int     i,c;
    float   x,y;

    if (fgets(buf,139,s)==NULL)
        return(0);
    for (i=0;buf[i]!='\n' && buf[i]!='\r' && buf[i]!=EOS;i++);
    buf[i]=EOS;
    if (sscanf(buf,"fpo pd %f %f",&x,&y)==2 && x>1 && x<15 && y>1 && y<15)
        {
        dp->pagex=x;
        dp->pagey=y;
        return(1);
        }
    if (sscanf(buf,"/cy %f %f",&x,&y)==2 && x!=y)
        {
        dp->ymin=x;
        dp->ymax=y;
        return(1);
        }
    if (sscanf(buf,"/cx %f %f",&x,&y)==2 && x!=y)
        {
        dp->xmin=x;
        dp->xmax=y;
        return(1);
        }
    if (buf[0]!='e' || buf[1]!='t')
        return(1);
    for (i=2;buf[i]==' ' || buf[i]=='\t';i++);
    c=buf[i];
    if (buf[i]!='y' && buf[i]!='x' && buf[i]!='z' && buf[i]!='g')
        return(1);
    for (i++;buf[i]==' ' || buf[i]=='\t';i++);
    if (buf[i]=='\"')
        i++;
    strcpy(buf2,&buf[i]);
    i=strlen(buf2);
    if (i-1>=0 && buf2[i-1]=='\"')
        buf2[i-1]=EOS;
    if (strlen(buf2)>98)
        buf2[98]=EOS;
    if (c=='x')
        strcpy(dp->xtitle,buf2);
    else if (c=='y')
        strcpy(dp->ytitle,buf2);
    else if (c=='z')
        strcpy(dp->ztitle,buf2);
    else
        strcpy(dp->gtitle,buf2);
    return(1);
    }


static STATUS premature(FILE *s)

    {
    fclose(s);
    error("Error!");
    normal("\r\n");
    error("Premature EOF!  Operation aborted!");
    shot_loaded=FALSE;
    normal("\r\n\r\n");
    wait_enter();
    return(ERROR);
    }


static int numgraphs(FILE *f)

    {
    int     ng;
    char    buf[10];

    fseek(f,0L,0);
    graphptr[0]=0;
    for (ng=1;1;ng++)
        {
        if (ng>1 && (ng&3)==1)
            normal("\r\n");
        sprintf(buf,"%2d. ",ng);
        hilite(buf);
        if (!goto_next_graph(f,0))
            break;
        if (ng<MAXGRAPHS)
            graphptr[ng]=ftell(f);
        }
    ngraphs=ng;
    return(ng);
    }


static int goto_next_graph(FILE *f,int quiet)

    {
    int          i,j,c;
    char         buf[20];
    char         buf2[20];
    int          ptrs[10];
    static char *pats[]={"\n//nc","\r//nc","\n/et g \"","\r/et g \"",""};
    long ptr;

    for (i=0;pats[i][0]!=EOS;i++)
        ptrs[i]=0;
    while ((c=fgetc(f))!=EOF)
        for (i=0;pats[i][0]!=EOS;i++)
            if (c==pats[i][ptrs[i]])
                {
                ptrs[i]++;
                if (pats[i][ptrs[i]]==EOS)
                    {
                    if (i==0 || i==1)
                        return(ptrs[i]);
                    if (!quiet)
                        {
                        ptr=ftell(f);
                        fread(buf,1,15,f);
                        fseek(f,ptr,0);
                        for (j=0;j<14 && buf[j]!='\n' && buf[j]!='\r'
                                      && buf[j]!='\"' && buf[j]!=EOS;j++);
                        buf[j]=EOS;
                        sprintf(buf2,"%-14s ",buf);
                        normal(buf2);
                        }
                    ptrs[i]=0;
                    }
                }
            else
                ptrs[i]=0;
    return(0);
    }


static void shotsize(FILE *f,int *x,int *y,int *bin,long maxpos)

    {
    char    buf[256];
    unsigned char a[4];
    unsigned int  xx;
    int     xo,xc,yo,xset,s,i,n,binary,c;
    long    startpos,p,len,size,pos;
    float   sq;

    (*x)=-1;
    (*y)=-1;
    startpos=ftell(f);
    len=maxpos-startpos;
    binary=0;
    (*bin)=0;
    do
        {
        p=ftell(f);
        n=fread(buf,sizeof(char),(size_t)(maxpos-p > 256 ? 256 : maxpos-p),f);
        for (i=0;i<n;i++)
            if (buf[i]<9 || (buf[i]>13 && buf[i]<32 && buf[i]!=26) || buf[i]>127)
                binary=1;
        } while (n==256);
    if (binary)
        {
        (*bin)=1;
        hilite("Binary ");
        sq=sqrt((float)len);
        s=sq;
        if (s!=sq)
            {
            fseek(f,startpos,0);
            fread(a,1,4,f);
            xx=a[0];
            xx= (xx<<8) | a[1];
            (*x)=(signed)xx;
            xx=a[2];
            xx= (xx<<8) | a[3];
            (*y)=(signed)xx;
            size=(long)(*x)*(long)(*y);
            if (size+4L!=len)
                {
                normal("\r\n");
                error("File dimensions cannot be determined.");
                normal("\r\n");
                (*x)=(*y)=-1;
                return;
                }
            }
        else
            (*x)=(*y)=s;
        if (!s)
            {
            normal("\r\n");
            error("File has no data.");
            normal("\r\n\r\n");
            return;
            }
        sprintf(buf,"%d x %d",(*x),(*y));
        hilite(buf);
        normal("\r\n");
        return;
        }
    hilite("ASCII ");
    fseek(f,startpos,0);
    pos=startpos;
    xset=0;
    xc=0;
    yo=0;
    while (1)
        {
        xc=0;
        while (1)
            {
            while (pos<maxpos && ((c=fgetc(f)),++pos,c!='\n')
                   && c!='e' && c!='E'
                   && (c<'0' || c>'9') && c!='.' && c!='-'
                   &&  c!='+' && c!=EOF && c!='/' && c!=';');
            if (pos>=maxpos || c==EOF)
                break;
            if (c=='\n')
                break;
            if (c=='/' || c==';')
                {
                while(pos<maxpos && ((c=fgetc(f)),++pos,c!='\n') && c!=EOF);
                if (pos>=maxpos || c==EOF)
                    break;
                xc=0;
                continue;
                }
            while (((c=fgetc(f)),++pos,c=='e') || c=='E' || (c>='0' && c<='9')
                   || c=='.' || c=='-' || c=='+');
            xc++;
            }
        if (!xc)
            if (pos>=maxpos || c==EOF)
                break;
            else
                continue;
        if (!xset)
            {
            xset=1;
            xo=xc;
            yo++;
            if (pos>=maxpos || c==EOF)
                break;
            else
                continue;
            }
        if (xo!=xc)
            {
            normal("\r\n");
            error("File is nonuniform.");
            normal("\r\n\r\n");
            return;
            }
        yo++;
        if (pos>=maxpos || c==EOF)
            break;
        }
    if (!yo || !xset)
        {
        normal("\r\n");
        error("File has no data points.");
        normal("\r\n\r\n");
        return;
        }
    (*x)=xo;
    (*y)=yo;
    fseek(f,0L,0);
    sprintf(buf,"%d x %d",xo,yo);
    hilite(buf);
    normal("\r\n");
    }


static BOOLEAN alloc_mem(DATAPTR d)

    {
    int             i;
    long            totalmem;
    long            nmainptrs;
    long            mainptrmem;
    long            nmainarray;
    long            mainarraymem;
    long            ninterpptrs;
    long            interpptrmem;
    long            ninterparray;
    long            interparraymem;
    long            stackmem;
    long            nstack;
    long            donemem;
    long            ndone;
    long            cf;
    void       huge*marker;

    if (d->memory!=NULL)
        {
        farfree(d->memory);
        d->memory=NULL;
        }
    nmainptrs=(long)d->xsize;
    mainptrmem=nmainptrs*sizeof(float far *);
    nmainarray=(long)d->xsize*(long)d->ysize;
    mainarraymem=nmainarray*sizeof(float);
    ninterpptrs=(long)d->xsize-1;
    interpptrmem=ninterpptrs*sizeof(float far *);
    ninterparray=(long)(d->xsize-1)*(long)(d->ysize-1);
    interparraymem=ninterparray*sizeof(float);
    nstack=(long)d->xsize*(long)d->ysize+10000L;
    stackmem=nstack*sizeof(int);
    ndone=(long)(d->xsize)*(long)(d->ysize)+10L;
    donemem=ndone*sizeof(char);

    totalmem=mainptrmem+mainarraymem+interpptrmem+interparraymem+
             stackmem+donemem+10000L;
    cf=farcoreleft();

    if (cf<totalmem||(d->memory=farmalloc(totalmem))==NULL)
        {
        error("There is not enough far memory to load this data file.");
        cprintf("\r\nYou have %ld free far bytes, and you need "
                "at least %ld.\r\n\7",cf,totalmem,totalmem-cf);
        return(FALSE);
        }
    else
        {
        cprintf("Memory allocation = %ldK out of %ldK available.\r\n",
                (totalmem+512L)>>10,(cf+512L)>>10);
        cprintf("Starting memory pointer = %Fp\r\n",d->memory);
        }

    /* Declare pointer to array of main 2D array pointers */
    d->shotdata=(float far*far*)d->memory;

    /* Declare array of pointers to main 2D array data */
    d->shotdata[0]=(float far*)((float huge*)d->shotdata+nmainptrs);
    for (i=1;i<d->xsize;i++)
        d->shotdata[i]=(float far*)((float huge*)d->shotdata[i-1]+d->ysize);
    marker=(void huge*)((float huge*)d->shotdata[d->xsize-1]+d->ysize);
    if (d->xsize>1 && d->ysize>1)
        {

        /* Declare pointer to array of interpolation array pointers */
        d->interp=(float far*far*)marker;

        /* Declare array of pointers to interpolation array data */
        d->interp[0]=(float far*)((float huge*)d->interp+ninterpptrs);
        for (i=1;i<d->xsize-1;i++)
            d->interp[i]=(float far*)((float huge*)d->interp[i-1]+d->ysize-1);

        marker=(void huge*)((float huge*)d->interp[d->xsize-2]+d->ysize-1);
        }

    /* Declare stack array pointer */
    d->stack=(int huge*)marker;

    /* Declare "done" array pointer */
    d->done=(unsigned char far*)((int huge*)d->stack+nstack);
    d->stacksize=nstack;
    d->donesize=ndone;
    return(TRUE);
    }
