/*
**
** PSHELP.C     PostScript help routines for CONTOUR.C
**
** Will Menninger, 12-15-92
**
*/

#include "contour.h"
#include <dir.h>
#include <dos.h>
#include <ps.h>

#define BOXLWIDTH   0.96
#define GLHWIDTH    0.48
#define GLLWIDTH    0.24
#define TICKWIDTH   0.96
#define GDL1        .5
#define GDL2        2.5
#define GDH1        1.
#define GDH2        2.

static float    gr_x1,gr_y1,gr_x2,gr_y2;
static float    px_x1,px_y1,px_x2,px_y2;
static float    yoff,yscale,xoff,xscale;
static int      ix1,iy1,ix2,iy2;
static float    yppu,xppu;
static char     xaxformat[20];
static char     yaxformat[20];
static float    fs;
static float    axtick;
static float    axtmar;
static int      axmodx,axmody;

static int extnum(char *name);


int nextname(int color,char *s,int *append,int *eject)

    {
    struct ffblk    ff;
    int             done;
    int             max;
    int             n;
    int             in;
    int             ask_append;
    char            buf[100];

    ask_append=0;
    ps_update(0,color,"Print to <F> File, <1>-<4> LPT1-4, <5>-<8> COM1-4,"
                      " <P> PRN, <ESC> Cancel",NULL);
    (*append)=0;
    (*eject)=1;
    while (1)
        {
        in=bbioskey(0)&0xff;
        if (in==27)
            {
            ps_clr_update();
            return(0);
            }
        if (in=='p' || in=='P')
            {
            strcpy(s,"PRN");
            ps_clr_update();
            return(1);
            }
        if (in>='1' && in<='4')
            {
            sprintf(s,"LPT%c",in);
            ps_clr_update();
            return(1);
            }
        if (in>='5' && in<='8')
            {
            sprintf(s,"COM%c",in-4);
            ps_clr_update();
            return(1);
            }
        if (in=='f' || in=='F')
            {
            ask_append=1;
            break;
            }
        }
    ps_clr_update();
    max=0;
    done=findfirst("out*.ps",&ff,0);
    while (!done)
        {
        n=extnum(ff.ff_name);
        if (n>max)
            max=n;
        done=findnext(&ff);
        }
    if (max && ask_append)
        {
        ps_clr_update();
        sprintf(buf,"Append output to file OUT%d.PS?  (y/n)",max);
        ps_update(0,color,buf,NULL);
        while (1)
            {
            in=bbioskey(0)&0xff;
            if (in==27)
                {
                ps_clr_update();
                return(0);
                }
            if (in=='n' || in=='N' || in=='y' || in=='Y')
                break;
            }
        ps_clr_update();
        (*append)=(in=='y' || in=='Y');
        }
    else
        (*append)=0;
    ps_clr_update();
    ps_update(0,color,"Send page eject code after plot?  (y/n)",NULL);
    while (1)
        {
        in=bbioskey(0)&0xff;
        if (in==27)
            {
            ps_clr_update();
            return(0);
            }
        if (in=='n' || in=='N' || in=='y' || in=='Y')
            break;
        }
    ps_clr_update();
    (*eject)=(in=='y' || in=='Y');
    sprintf(s,"OUT%d.PS",(*append) ? max : max+1);
    return(1);
    }


void ps_delete(void)

    {
    struct ffblk    ff;
    int             done;

    done=findfirst("out*.ps",&ff,0);
    while (!done)
        {
        remove(ff.ff_name);
        done=findnext(&ff);
        }
    }


int fontsize(float px,float py)

    {
    float   x;

    x= (px>py) ? py : px;
    if (x<=2)
        return(10);
    else if (x<=4)
        return(12);
    else if (x<=5)
        return(14);
    else if (x<=6)
        return(16);
    else
        return(18);
    }


void finalprint(char *filename,int okay)

    {
    FILE   *f;
    long    len;
    char    buf[100];

    if (tolower(filename[0])!='o')
        len=-1;
    else
        {
        f=fopen(filename,"rb");
        if (f==NULL)
            len=-1;
        else
            {
            fseek(f,0L,2);
            len=ftell(f);
            fclose(f);
            len=(len+512)>>10;
            }
        }
    if (len<0)
        sprintf(buf,"Postscript%s written to %s",
                okay ? "":" partially",filename);
    else if (len==0)
        sprintf(buf,"PostScript file%s written to %s (<1K)",
                okay ? "": " partially",filename);
    else
        sprintf(buf,"PostScript file%s written to %s (%ldK)",
                okay ? "": " partially",filename,len);
    ps_update(0,14,buf,NULL);
    }


/*
** (x1,y1) = lower left corner, (x2,y2)=upper right corner
** (px1,py1) = lower left PostScript corner in points.
** (px2,py2) = upper right PostScript corner in points.
**
*/
void ps_setgriddim(float x1,float y1,float x2,float y2,
                   float px1,float py1,float px2,float py2,
                   float fsize,float axtck,float axtm,
                   char *xf,char *yf)

    {
    axtick=axtck;
    axtmar=axtm;
    fs=fsize;
    gr_x1=x1;
    gr_y1=y1;
    gr_x2=x2;
    gr_y2=y2;
    px_x1=px1;
    px_x2=px2;
    px_y1=py1;
    px_y2=py2;
    strcpy(xaxformat,xf);
    strcpy(yaxformat,yf);
    yppu=(px_y2-px_y1)/(gr_y2-gr_y1);
    xppu=(px_x2-px_x1)/(gr_x2-gr_x1);
    getticks(gr_x1,gr_x2,&xoff,&xscale,&ix1,&ix2,&axmodx,15);
    getticks(gr_y1,gr_y2,&yoff,&yscale,&iy1,&iy2,&axmody,15);
    }


void ps_bbox(void)

    {
    g_grey(0.);
    g_linecap(2);
    g_linejoin(0);
    g_linewidth(BOXLWIDTH);
    g_moveto(px_x1,px_y1);
    g_lineto(px_x1,px_y2);
    g_lineto(px_x2,px_y2);
    g_lineto(px_x2,px_y1);
    g_closepath();
    g_stroke();
    }


void ps_putgrid(int grids)

    {
    int     i;
    float   xval,pxval;
    char    buf[100];

    g_linecap(2);
    g_linejoin(0);
    g_grey(0.);
    g_justify(5);
    for (i=iy1;i<=iy2;i++)
        {
        xval=yoff+i*yscale;
        if (fabs(xval/yscale)<1e-4)
            xval=0;
        pxval=px_y1+(xval-gr_y1)*yppu;
        if (i%axmody==0)
            {
            if (grids)
                {
                g_dash(GDH1,GDH2);
                g_linewidth(GLHWIDTH);
                g_line(px_x1,pxval,px_x2,pxval);
                }
            g_dash(0.,0.);
            g_linewidth(TICKWIDTH);
            g_line(px_x1,pxval,px_x1-axtick,pxval);
            sprintf(buf,yaxformat,xval);
            g_moveto(px_x1-axtick-axtmar,pxval);
            g_showtext(buf);
            }
        else
            {
            g_dash(0.,0.);
            g_linewidth(TICKWIDTH);
            g_line(px_x1,pxval,px_x1-axtick/2.,pxval);
            if (grids)
                {
                g_linewidth(GLLWIDTH);
                g_dash(GDL1,GDL2);
                g_line(px_x1,pxval,px_x2,pxval);
                }
            }
        }
    g_justify(1);
    for (i=ix1;i<=ix2;i++)
        {
        xval=xoff+i*xscale;
        if (fabs(xval/xscale)<1e-4)
            xval=0;
        pxval=px_x1+(xval-gr_x1)*xppu;
        if (i%axmodx==0)
            {
            if (grids)
                {
                g_linewidth(GLHWIDTH);
                g_dash(GDH1,GDH2);
                g_line(pxval,px_y1,pxval,px_y2);
                }
            g_linewidth(TICKWIDTH);
            g_dash(0.,0.);
            g_line(pxval,px_y1-axtick,pxval,px_y1);
            sprintf(buf,xaxformat,xval);
            g_moveto(pxval,px_y1-axtick-axtmar);
            g_showtext(buf);
            }
        else
            {
            g_linewidth(TICKWIDTH);
            g_dash(0.,0.);
            g_line(pxval,px_y1-axtick/2.,pxval,px_y1);
            if (grids)
                {
                g_linewidth(GLLWIDTH);
                g_dash(GDL1,GDL2);
                g_line(pxval,px_y1,pxval,px_y2);
                }
            }
        }
    g_dash(0.,0.);
    }


void ps_put_titles(DATAPTR d,float xt,float yt,float xx,float yx,
                             float xy,float yy)

    {
    g_grey(0.);
    g_justify(7);
    g_moveto(xt,yt);
    g_showtext(d->gtitle);
    g_justify(1);
    g_moveto(xx,yx);
    g_showtext(d->xtitle);
    g_justify(7);
    g_moveto(xy,yy);
    g_rotate(90.);
    g_showtext(d->ytitle);
    g_rotate(270.);
    }


void ps_update(int prompt,int color,char *message,void *ptr)

    {
    float  *a;
    int    *j;
    int     i;
    int     type;
    char    buf[100];

    ps_clr_update();
    settextjustify(LEFT_TEXT,TOP_TEXT);
    type=0;
    for (i=0;message[i]!=EOS;i++)
        if (message[i]=='%')
            {
            for (i++;message[i]!=EOS && message[i]!='%';i++)
                if (message[i]=='g' || message[i]=='f')
                    {
                    type=1;
                    break;
                    }
                else if (message[i]=='d' || message[i]=='u')
                    {
                    type=2;
                    break;
                    }
                else if (message[i]=='s')
                    {
                    type=3;
                    break;
                    }
            if (type)
                break;
            }
    if (type==1)
        {
        a=(float *)ptr;
        sprintf(buf,message,(*a));
        }
    else if (type==2)
        {
        j=(int *)ptr;
        sprintf(buf,message,(*j));
        }
    else if (type==3)
        sprintf(buf,message,ptr);
    else
        strcpy(buf,message);
    if (prompt)
        strcat(buf," <PRESS ANY KEY>");
    setcolor(color);
    outtextxy(0,0,buf);
    if (prompt)
        {
        bbioskey(0);
        ps_clr_update();
        }
    }


void ps_clr_update(void)

    {
    setfillstyle(SOLID_FILL,BLACK);
    bar(0,0,639,8);
    }


static int extnum(char *name)

    {
    int     i,j;

    for (i=3;name[i]!='.';i++)
        if (name[i]<'0' || name[i]>'9')
            return(-1);
    j=0;
    for (i=3;name[i]!='.';i++)
        j=j*10+name[i]-'0';
    return(j);
    }
