/*
**
** PSIPLOT.C   Does an intensity plot in PostScript
**
** Will Menninger, 12/92, programmed in Turbo C 2.0
**
**
*/

#include "contour.h"
#include <ps.h>

#define LTHICK      .96
#define AXTICK      0.012
#define AXTMAR      0.008
#define YAXNMAR     4.
#define XAXNMAR     .5
#define TMAR        1.5
#define LNMAR1      4.
#define LNMAR2      0.3
#define LEGWIDTH    18.

#define NDIVLEG     1000

#define XL1     57
#define XL2     70
#define XB      140
#define BOXH    120
#define BOXW    480

#define PS_COLOR_RES    16

static int pscolor;
static float axtick;
static float axtmar;

static void def_colors(int res,int dark_to_light);
static void def_grey(int dark_to_light);


void ps_iplot(char *filename,DATAPTR d,float zmin,float zmax,int grids,
              int color,int dark_to_light,int append,int eject)

    {
    int             i,j,x,y;
    int             i1,i2,modval;
    float           offset,scale;
    char            buf[100];
    float           fy,z,zscale;
    float           totalheight,totalwidth;
    float           psleft,psright,pstop,psbottom;
    float           pslx,psx1,psy1,psx2,psy2;
    float           dmin,pagex,pagey,fs,f;

    pscolor=color;
    zscale=zmax-zmin;
    if (d->min>=0.)
        dmin=0.;
    else
        dmin=d->min;
    if (d->pagex<0)
        {
        pagex=5.0;
        pagey=5.0;
        }
    else
        {
        pagex=d->pagex;
        pagey=d->pagey;
        }
    ps_update(0,15,"Writing PostScript to %s ...",filename);
    fs=fontsize(pagex,pagey);
    axtick=AXTICK*(pagex+pagey)*36.;
    axtmar=AXTMAR*(pagex+pagey)*36.;
    totalheight=pagey*72.+(3.+TMAR+XAXNMAR)*fs+axtick+axtmar;
    totalwidth=pagex*72.+LEGWIDTH+(YAXNMAR+LNMAR1+LNMAR2+1.)*fs+2.*(axtick+axtmar);
    psbottom=d->pagebotmar<-90. ? 11.*36.-totalheight/2. : d->pagebotmar*72.;
    pstop=psbottom+totalheight;
    psleft=d->pageleftmar<-90. ? 8.5*36.-totalwidth/2. : d->pageleftmar*72.;
    psright=psleft+totalwidth;
    psx1=psright-pagex*72.;
    psy1=psbottom+(2.+XAXNMAR)*fs+axtick+axtmar;
    psx2=psright;
    psy2=psy1+pagey*72.;
    pslx=psleft+fs*LNMAR1+axtick+axtmar;
    ps_setgriddim(d->xmin-.5*(d->xmax-d->xmin)/(d->xsize-1.),
                  d->ymin-.5*(d->ymax-d->ymin)/(d->ysize-1.),
                  d->xmax+.5*(d->xmax-d->xmin)/(d->xsize-1.),
                  d->ymax+.5*(d->ymax-d->ymin)/(d->ysize-1.),
                  psx1,psy1,psx2,psy2,
                  fs,axtick,axtmar,"%g","%g");
    if (append)
        g_append(filename);
    else
        g_init(filename);
    if (pscolor)
        def_colors(PS_COLOR_RES,dark_to_light);
    else
        def_grey(dark_to_light);
    sprintf(buf,"\\fTR\\s{%g}",fs);
    g_showtext(buf);

    /* Draw titles */

    ps_put_titles(d,(psleft+psright)/2.,pstop+fs,(psx1+psx2)/2.,psbottom+fs,
                    psx1-axtmar-YAXNMAR*fs,(psy1+psy2)/2.);
    g_justify(6);
    g_moveto(psleft,psy2+fs);
    g_showtext(d->ztitle);

    /* Output function that will do the legend */

    sprintf(buf,"0 1 %d \n",(int)(NDIVLEG-1));
    g_psout(buf);
    sprintf(buf," { dup %d div %g mul %g add %g sub %g div\n",
             (int)(NDIVLEG),d->max-dmin,dmin,zmin,zmax-zmin);
    g_psout(buf);
    sprintf(buf,"dup floor sub setshade %g mul neg %g add %g\n",
             (psy2-psy1)/NDIVLEG,psy2,pslx);
    g_psout(buf);
    sprintf(buf,"exch moveto 0 -%g rlineto %g 0 rlineto\n",
            (psy2-psy1)/NDIVLEG,(float)LEGWIDTH);
    g_psout(buf);
    sprintf(buf,"0 %g rlineto closepath fill } for\n",(psy2-psy1)/NDIVLEG);
    g_psout(buf);

    /* Box around the legend */

    g_linewidth(LTHICK);
    g_linecap(2);
    g_linejoin(0);
    g_grey(0.);
    g_moveto(pslx,psy1);
    g_lineto(pslx,psy2);
    g_lineto(pslx+LEGWIDTH,psy2);
    g_lineto(pslx+LEGWIDTH,psy1);
    g_closepath();
    g_stroke();

    /* Mark the axis of the legend with numbers */

    g_justify(5);
    getticks(dmin,d->max,&offset,&scale,&i1,&i2,&modval,25);
    for (i=i1;i<=i2;i++)
        {
        if (i%modval)
            continue;
        z=offset+i*scale;
        if (fabs(z/scale)<1e-4)
            z=0;
        y=psy2-(z-dmin)*(psy2-psy1)/(d->max-dmin);
        g_line(pslx,y,pslx-axtick,y);
        g_moveto(pslx-axtick-axtmar,y);
        sprintf(buf,"%g",z);
        g_showtext(buf);
        }


    /* Define the PS function that will draw squares */

    g_psout("/S {/CNT exch def /Y exch def /X exch def\n");
    g_psout("CNT 1 sub -1 0\n");
    sprintf(buf,"{ X add %g mul %g add Y %g mul %g add moveto\n",
            (psx2-psx1)/d->xsize,psx1,(psy2-psy1)/d->ysize,psy1);
    g_psout(buf);
    sprintf(buf," setshade 0 %g rlineto %g 0 rlineto 0 -%g rlineto\n",
            (psy2-psy1)/d->ysize,(psx2-psx1)/d->xsize,
            (psy2-psy1)/d->ysize);
    g_psout(buf);
    g_psout(" closepath fill } for } def\n");

    for (y=0;y<d->ysize;y++)
        {
        if (bbioskey(1))
            break;
        if (y%5==0)
            {
            fy=d->ymin+y*(d->ymax-d->ymin)/(d->ysize-1.);
            ps_update(0,15,"Y = %g",&fy);
            }
        for (x=0;x<d->xsize;x+=10)
            {
            if (x+9>d->xsize-1)
                i=d->xsize-x;
            else
                i=10;
            if (!i)
                continue;
            for (j=0;j<i;j++)
                {
                f=(d->shotdata[x+j][y]-zmin)/zscale;
                if (f==floor(f))
                    if (f==0.)
                        f=0.;
                    else
                        f=1.;
                else
                    f=f-floor(f);
                sprintf(buf,"%.3f ",f);
                g_psout(buf);
                }
            sprintf(buf,"%d %d %d S\n",x,y,i);
            g_psout(buf);
            }
        }


    /* Bounding rectangle */

    ps_bbox();


    /* Grids */

    ps_putgrid(grids);

    if (eject)
        g_showpage();

    g_close();
    if (bbioskey(1))
        {
        bbioskey(0);
        ps_update(1,12,"PostScript graph interrupted.",NULL);
        finalprint(filename,0);
        }
    else
        finalprint(filename,1);
    }


static void def_colors(int res,int dark_to_light)

    {
    static float d2l[7][4]={{.5,.5,.5,6},{0,0,0,1},{0,0,1,2},{1,0,0,3},
                            {0,1,0,3},{1,1,0,3},{1,1,1,3}};
    static float l2d[7][4]={{.5,.5,.5,6},{1,1,1,1},{1,1,0,3},{0,1,0,3},
                            {1,0,0,3},{0,0,1,3},{0,0,0,2}};
    static char *arname[] = {"redarray","greenarray","bluearray"};
    int          colorgun,i,n,j,arraysize;
    float        val;
    char         buf[80];
    float        cmap[7][4];

    if (dark_to_light)
        for (i=0;i<=d2l[0][3];i++)
            for (j=0;j<4;j++)
                cmap[i][j]=d2l[i][j];
    else
        for (i=0;i<=l2d[0][3];i++)
            for (j=0;j<4;j++)
                cmap[i][j]=l2d[i][j];

    for (colorgun=0;colorgun<3;colorgun++)
        {
        arraysize=0;
        sprintf(buf,"/%s\n[",arname[colorgun]);
        g_psout(buf);
        for (i=1;i<cmap[0][3];i++)
            {
            n=res*cmap[i+1][3];
            for (j=0;j<n;j++)
                {
                val=cmap[i][colorgun]
                   +(cmap[i+1][colorgun]-cmap[i][colorgun])*(float)j/n;
                sprintf(buf,"%4.3f",val);
                g_psout(buf);
                arraysize++;
                if (arraysize%12==0)
                    g_psout("\n");
                else
                    g_psout(" ");
                }
            }
        sprintf(buf,"%4.3f] def\n",cmap[i][colorgun]);
        g_psout(buf);
        arraysize++;
        }
    sprintf(buf,"/setshade {%d mul cvi /arind exch def\n",arraysize-1);
    g_psout(buf);
    sprintf(buf,"           %s arind get %s arind get %s arind get\n",
                arname[0],arname[1],arname[2]);
    g_psout(buf);
    g_psout("           setrgbcolor} def\n");
    }


static void def_grey(int dark_to_light)

    {
    if (dark_to_light)
        g_psout("/setshade {setgray} def\n");
    else
        g_psout("/setshade {1.0 sub neg setgray} def\n");
    }
