/*
**
** SAVEPCX.C    Saves a PCX format bitmap.
**
** Will Menninger, July 1989, updated 3/95
**
*/

#include "contour.h"

#define    PUTWORD(x)  if (!fputword((x),stream)) return(ERROR)
#define    PUTCHAR(x)  if (fputc((x),stream)==EOF) return(ERROR)

#define     CONTROLLER      0x3ce
#define     REGISTER        0x3cf
#define     REG_SELECT      0x3c4
#define     DATA_REG        0x3c5
#define     MAP_MASK        2
#define     EGA_SEG         0xa000


static int  writevgaline  (FILE *stream,int n);
static int  writepair     (FILE *stream,int bytecount,int cbyte);
static int  write_opening (FILE *stream);
static int  fputword      (int w,FILE *stream);
static void readvgapal    (int pen,int *r,int *g,int *b);
static int  getline       (int n,char *buf);
static void getegamem(char *buf,int x1,int y1,int x2,int y2,int bp1,int bp2,
                      int bw);

static int picnum=0;
static char report[80];


char *vgasave(void)

   {
   FILE *f;
   char filename[80];
   int     i,status;

   sprintf(filename,"pic%03d.pcx",picnum++);
   f=fopen(filename,"wb");
   sprintf(report,"Error saving picture as %s.",filename);
   if (f==NULL)
       return(report);
   if (write_opening(f)==ERROR)
       {
       fclose(f);
       return(report);
       }
   for (i=0;i<480;i++)
       {
       status=writevgaline(f,i);
       if (status==ERROR)
           {
           fclose(f);
           return(report);
           }
       }
   fclose(f);
   sprintf(report,"Picture successfully saved as %s.",filename);
   return(report);
   }


static int writevgaline(FILE *stream,int n)

   {
   char    linebuf[512];
   int     bytes,i,bytecount,cbyte;

   bytes=getline(n,linebuf);
   cbyte=linebuf[0];
   bytecount=1;
   for (i=1;i<bytes;i++)
       if (linebuf[i]!=cbyte)
           {
           if (writepair(stream,bytecount,cbyte)==ERROR)
               return(ERROR);
           bytecount=1;
           cbyte=linebuf[i];
           continue;
           }
       else
           {
           bytecount++;
           if (bytecount==64)
               {
               if (writepair(stream,63,cbyte)==ERROR)
                   return(ERROR);
               bytecount=1;
               }
           }
   return(writepair(stream,bytecount,cbyte));
   }


static int writepair(FILE *stream,int bytecount,int cbyte)

   {
   if (!bytecount)
       return(NO_ERROR);
   if (bytecount>1 || (cbyte & 0xc0)==0xc0)
       if (fputc(bytecount | 0xc0,stream)==EOF)
           return(ERROR);
   if (fputc(cbyte,stream)==EOF)
        return(ERROR);
   return(NO_ERROR);
   }


static int write_opening(FILE *stream)

   {
   int     i,r,g,b;
   int     w,h,bp,bw;

   w=640;
   h=480;
   bp=4;
   bw=80;
   fseek(stream,0L,0);
   PUTWORD(0x50a);
   PUTWORD(0x101);
   PUTWORD(0);
   PUTWORD(0);
   PUTWORD(w-1);
   PUTWORD(h-1);
   PUTWORD(0x4b);
   PUTWORD(0x4b);
   for (i=0;i<16;i++)
       {
       readvgapal(i,&r,&g,&b);
       PUTCHAR(r<<2);
       PUTCHAR(g<<2);
       PUTCHAR(b<<2);
       }
   PUTCHAR(0);
   PUTCHAR(bp);
   PUTWORD(bw);
   PUTWORD(0);
   PUTWORD(h);
   for (i=0;i<28;i++)
       PUTWORD(0);
   return(NO_ERROR);
   }


static int fputword(int w,FILE *stream)

   {
   if (fputc(w&0xff,stream)==EOF)
       return(0);
   if (fputc((w>>8)&0xff,stream)==EOF)
       return(0);
   return(1);
   }


static void readvgapal(int pen,int *r,int *g,int *b)

    {
    union REGS      set;

    set.h.bl=pen;
    set.x.ax=0x1007;
    int86(0x10,&set,&set);
    set.h.bl=set.h.bh;
    set.h.bh=0;
    set.x.ax=0x1015;
    int86(0x10,&set,&set);
    (*r)=set.h.dh;
    (*g)=set.h.ch;
    (*b)=set.h.cl;
    }


static int getline(int n,char *buf)

   {
   getegamem(buf,0,n,79,n,0,3,80);
   return(320);
   }


static void getegamem(char *buf,int x1,int y1,int x2,int y2,int bp1,int bp2,
                      int bw)

    {
    int     p,i,j,y,memoff;

    p=0;
    for (i=bp1;i<=bp2;i++)
        {
        outportb(CONTROLLER,4);
        outportb(REGISTER,i);
        for (y=y1;y<=y2;y++)
            {
            memoff=y*bw;
            for (j=x1;j<=x2;j++)
                buf[p++]=peekb(EGA_SEG,memoff+j);
            }
        }
    }
